import asyncio
import copy
import logging
import warnings
from collections import defaultdict
from typing import Dict, List, Optional

from hummingbot.connector.derivative.position import Position
from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.data_type.common import PositionMode, PositionSide
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class PerpetualTrading:
    """Keeps perpetual trading state."""

    _logger: Optional[HummingbotLogger] = None

    def __init__(self, trading_pairs: List[str]):
        self._account_positions: Dict[str, Position] = {}
        self._position_mode: PositionMode = PositionMode.ONEWAY
        self._leverage: Dict[str, int] = defaultdict(lambda: 1)
        self._trading_pairs = trading_pairs

        self._funding_info: Dict[str, FundingInfo] = {}
        self._funding_payment_span: List[int] = [0, 0]
        self._funding_info_stream = asyncio.Queue()

        self._funding_info_updater_task: Optional[asyncio.Task] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    @property
    def account_positions(self) -> Dict[str, Position]:
        """
        Returns a dictionary of current active open positions
        """
        return self._account_positions

    @property
    def funding_info(self) -> Dict[str, FundingInfo]:
        """
        The funding information per trading pair.
        """
        return copy.deepcopy(self._funding_info)

    @property
    def funding_info_stream(self) -> asyncio.Queue:
        """
        The stream to which to supply funding info updates to be processed by the class.
        """
        return self._funding_info_stream

    def set_position(self, pos_key: str, position: Position):
        self.logger().debug(f"Setting position {pos_key} to {Position}")
        self._account_positions[pos_key] = position

    def remove_position(self, post_key: str) -> Optional[Position]:
        return self._account_positions.pop(post_key, None)

    def initialize_funding_info(self, funding_info: FundingInfo):
        """
        Initializes a single trading pair funding information.
        """
        self._funding_info[funding_info.trading_pair] = funding_info

    def add_trading_pair(self, trading_pair: str):
        """
        Adds a trading pair to the list of tracked trading pairs.
        This should be called when dynamically adding a trading pair.
        Note: Funding info should be initialized separately via initialize_funding_info().

        :param trading_pair: the trading pair to add
        """
        if trading_pair not in self._trading_pairs:
            self._trading_pairs.append(trading_pair)

    def remove_trading_pair(self, trading_pair: str):
        """
        Removes a trading pair from the list of tracked trading pairs and cleans up related data.
        This should be called when dynamically removing a trading pair.

        :param trading_pair: the trading pair to remove
        """
        if trading_pair in self._trading_pairs:
            self._trading_pairs.remove(trading_pair)
        # Clean up funding info
        self._funding_info.pop(trading_pair, None)
        # Clean up leverage settings
        if trading_pair in self._leverage:
            del self._leverage[trading_pair]

    def is_funding_info_initialized(self) -> bool:
        """
        Checks if there is funding information for all trading pairs.
        """
        return all(
            trading_pair in self._funding_info
            for trading_pair in self._trading_pairs
        )

    def start(self):
        """
        Starts the async task that updates the funding information from the updates stream queue.
        """
        self.stop()
        self._funding_info_updater_task = safe_ensure_future(
            self._funding_info_updater()
        )

    def stop(self):
        """
        Stops the funding info updating async task.
        """
        if self._funding_info_updater_task is not None:
            self._funding_info_updater_task.cancel()
            self._funding_info_updater_task = None
        self._funding_info.clear()

    def position_key(self, trading_pair: str, side: PositionSide = None, mode: PositionMode = None) -> str:
        """
        Returns a key to a position in account_positions. On OneWay position mode this is the trading pair.
        On Hedge position mode this is a combination of trading pair and position side
        :param trading_pair: The market trading pair
        :param side: The position side (long or short)
        :return: A key to the position in account_positions dictionary
        """
        pos_key = ""
        if mode is not None:
            pos_key = f"{trading_pair}{side.name}" if mode == PositionMode.HEDGE else trading_pair
        else:
            pos_key = f"{trading_pair}{side.name}" if self._position_mode == PositionMode.HEDGE else trading_pair
        return pos_key

    def get_position(self, trading_pair: str, side: PositionSide = None) -> Optional[Position]:
        """
        Returns an active position if exists, otherwise returns None
        :param trading_pair: The market trading pair
        :param side: The position side (long or short)
        :return: A position from account_positions or None
        """
        return self.account_positions.get(self.position_key(trading_pair, side), None)

    @property
    def funding_payment_span(self) -> List[int]:
        """
        Time span(in seconds) before and after funding period when exchanges consider active positions eligible for
        funding payment.
        :return: a list of seconds (before and after)
        """
        return self._funding_payment_span

    @property
    def position_mode(self) -> PositionMode:
        return self._position_mode

    def set_position_mode(self, value: PositionMode):
        """
        Sets position mode for perpetual trading, a child class might need to override this to set position mode on
        the exchange
        :param value: the position mode
        """
        self._position_mode = value

    def get_leverage(self, trading_pair: str) -> int:
        """
        Gets leverage level of a particular market
        :param trading_pair: the market trading pair
        :return: leverage level
        """
        return self._leverage[trading_pair]

    def set_leverage(self, trading_pair: str, leverage: int = 1):
        """
        Sets leverage level, e.g. 2x, 10x, etc..
        A child class may need to override this to set leverage level on the exchange
        :param trading_pair: the market trading pair
        :param leverage: leverage to be used
        """
        self._leverage[trading_pair] = leverage

    def get_funding_info(self, trading_pair: str) -> FundingInfo:
        """
        Returns funding information
        :param trading_pair: the market trading pair
        :return: funding info
        """
        return self._funding_info[trading_pair]

    async def _funding_info_updater(self):
        while True:
            try:
                funding_info_message: FundingInfoUpdate = await self._funding_info_stream.get()
                trading_pair = funding_info_message.trading_pair
                if trading_pair not in self._funding_info:
                    # Skip updates for trading pairs that haven't been initialized yet
                    # This can happen when pairs are added dynamically
                    self.logger().debug(
                        f"Received funding info update for uninitialized pair {trading_pair}, skipping."
                    )
                    continue
                funding_info = self._funding_info[trading_pair]
                funding_info.update(funding_info_message)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error updating funding info.", exc_info=True)

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        warnings.warn(
            "This method is replaced by PerpetualDerivativePyBase.get_buy_collateral_token, and will be removed"
            " once all perpetual connectors are updated to the latest standards.",
            DeprecationWarning,
            stacklevel=2,
        )
        _, quote = split_hb_trading_pair(trading_pair)
        return quote

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        warnings.warn(
            "This method is replaced by PerpetualDerivativePyBase.get_sell_collateral_token, and will be removed"
            " once all perpetual connectors are updated to the latest standards.",
            DeprecationWarning,
            stacklevel=2,
        )
        _, quote = split_hb_trading_pair(trading_pair)
        return quote



import asyncio
from collections import defaultdict
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional

import pandas as pd

import hummingbot.connector.derivative.bitmart_perpetual.bitmart_perpetual_constants as CONSTANTS
import hummingbot.connector.derivative.bitmart_perpetual.bitmart_perpetual_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.bitmart_perpetual.bitmart_perpetual_derivative import (
        BitmartPerpetualDerivative,
    )


class BitmartPerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):
    _bpobds_logger: Optional[HummingbotLogger] = None
    _trading_pair_symbol_map: Dict[str, Mapping[str, str]] = {}
    _mapping_initialization_lock = asyncio.Lock()

    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(
            self,
            trading_pairs: List[str],
            connector: 'BitmartPerpetualDerivative',
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DOMAIN
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._trading_pairs: List[str] = trading_pairs
        self._message_queue: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._exchange_info_listener_task = safe_ensure_future(self.listen_for_exchange_info())
        self._trade_messages_queue_key = CONSTANTS.TRADE_STREAM_CHANNEL
        self._snapshot_messages_queue_key = CONSTANTS.ORDER_BOOK_CHANNEL + "_SNAPSHOT"
        self._diff_messages_queue_key = CONSTANTS.ORDER_BOOK_CHANNEL + "_DIFF"
        self._funding_info_messages_queue_key = CONSTANTS.FUNDING_INFO_CHANNEL
        self._tickers_messages_queue_key = CONSTANTS.TICKERS_CHANNEL
        self._last_index_prices: Dict[str, Decimal] = {}
        self._last_mark_prices: Dict[str, Decimal] = {}

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        symbol_response, funding_response = await asyncio.gather(
            self._request_complete_contract_details(trading_pair),
            self._request_complete_funding_info(trading_pair)
        )

        symbol_data = symbol_response["data"].get("symbols")
        funding_data = funding_response.get("data")

        if symbol_data is not None and funding_data is not None:
            # TODO: Check if last_price replaces mark_price
            funding_info = FundingInfo(
                trading_pair=trading_pair,
                index_price=Decimal(symbol_data[0].get("index_price")),
                mark_price=Decimal(symbol_data[0].get("last_price")),
                next_funding_utc_timestamp=int(float(funding_data.get("funding_time")) * 1e-3),
                rate=Decimal(funding_data.get("expected_rate"))
            )
            return funding_info

    async def _connected_websocket_assistant(self) -> WSAssistant:
        url = f"{web_utils.wss_url(CONSTANTS.PUBLIC_WS_ENDPOINT, domain=self._domain)}"
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            stream_id_channel_pairs = [
                CONSTANTS.ORDER_BOOK_CHANNEL,
                CONSTANTS.TRADE_STREAM_CHANNEL,
                CONSTANTS.FUNDING_INFO_CHANNEL,
                CONSTANTS.TICKERS_CHANNEL,
            ]
            for channel in stream_id_channel_pairs:
                params = []
                for trading_pair in self._trading_pairs:
                    symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                    params.append(f"{channel}{f':{symbol.upper()}' if channel != 'futures/ticker' else ''}")
                    payload = {
                        "action": "subscribe",
                        "args": params,
                    }
                    subscribe_request: WSJSONRequest = WSJSONRequest(payload)
                    await ws.send(subscribe_request)
            self.logger().info("Subscribed to public order book, trade and funding info channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if event_message.get("data") is not None:
            stream_name = event_message.get("group")
            if CONSTANTS.ORDER_BOOK_CHANNEL in stream_name:
                if event_message["data"]["type"] == "update":
                    channel = self._diff_messages_queue_key
                else:
                    channel = self._snapshot_messages_queue_key
            elif CONSTANTS.TRADE_STREAM_CHANNEL in stream_name:
                channel = self._trade_messages_queue_key
            elif CONSTANTS.FUNDING_INFO_CHANNEL in stream_name:
                channel = self._funding_info_messages_queue_key
            elif CONSTANTS.TICKERS_CHANNEL in stream_name:
                channel = self._tickers_messages_queue_key
        return channel

    def _get_messages_queue_keys(self) -> List[str]:
        return [
            self._snapshot_messages_queue_key,
            self._diff_messages_queue_key,
            self._trade_messages_queue_key,
            self._funding_info_messages_queue_key,
            self._tickers_messages_queue_key
        ]

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        if len(raw_message["data"]) > 0:
            trade_data = raw_message["data"][0]
            trade_data["symbol"] = await self._connector.trading_pair_associated_to_exchange_symbol(trade_data["symbol"])
            trade_data["created_at"] = pd.to_datetime(trade_data["created_at"]).timestamp()
            trade_message: OrderBookMessage = OrderBookMessage(
                OrderBookMessageType.TRADE,
                {
                    "trading_pair": trade_data["symbol"],
                    "trade_type": self._parse_trade_way(trade_data["way"]),
                    "trade_id": trade_data["trade_id"],
                    "price": trade_data["deal_price"],
                    "amount": trade_data["deal_vol"]
                },
                timestamp=trade_data["created_at"])
            message_queue.put_nowait(trade_message)

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_data: Dict[str, Any] = snapshot_response.get("data")
        snapshot_timestamp: float = snapshot_data["timestamp"] / 1e3
        snapshot_data.update({"trading_pair": trading_pair})
        snapshot_msg: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": snapshot_data["trading_pair"],
            "update_id": 1,
            "bids": [(bid[0], bid[1]) for bid in snapshot_data["bids"]],
            "asks": [(ask[0], ask[1]) for ask in snapshot_data["asks"]]
        }, timestamp=snapshot_timestamp)
        return snapshot_msg

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        raw_message["data"]["symbol"] = await self._connector.trading_pair_associated_to_exchange_symbol(
            raw_message["data"]["symbol"])
        data = raw_message["data"]
        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.DIFF, {
            "trading_pair": data["symbol"],
            "update_id": int(data["version"]),
            "bids": [(depth["price"], depth["vol"]) for depth in data["bids"]],
            "asks": [(depth["price"], depth["vol"]) for depth in data["asks"]],
        }, timestamp=data["ms_t"] / 1e3)
        message_queue.put_nowait(order_book_message)

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        raw_message["data"]["symbol"] = await self._connector.trading_pair_associated_to_exchange_symbol(
            raw_message["data"]["symbol"])
        data = raw_message["data"]
        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": data["symbol"],
            "update_id": int(data["version"]),
            "bids": [(depth["price"], depth["vol"]) for depth in data["bids"]],
            "asks": [(depth["price"], depth["vol"]) for depth in data["asks"]],
        }, timestamp=data["ms_t"] / 1e3)
        message_queue.put_nowait(order_book_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data: Dict[str, Any] = raw_message["data"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(data["symbol"])

        if trading_pair not in self._trading_pairs:
            return
        next_funding_utc_timestamp = data.get("nextFundingTime")
        rate = data.get("fundingRate")
        funding_info = FundingInfoUpdate(
            trading_pair=trading_pair,
            index_price=self._last_index_prices.get(trading_pair),
            mark_price=self._last_mark_prices.get(trading_pair),
            next_funding_utc_timestamp=(int(float(next_funding_utc_timestamp) * 1e-3)
                                        if next_funding_utc_timestamp is not None
                                        else None),
            rate=Decimal(rate) if rate is not None else None,
        )
        message_queue.put_nowait(funding_info)

    async def listen_for_exchange_info(self):
        """
        Reads the exchange info events queue. For each trading pair initialized stores in a local variable
        the index and last price

        :param ev_loop: the event loop the method will run in
        """
        message_queue = self._message_queue[self._tickers_messages_queue_key]
        while True:
            try:
                diff_event = await message_queue.get()
                await self._parse_exchange_info_message(raw_message=diff_event)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public order book updates from exchange")

    async def _parse_exchange_info_message(self, raw_message: Dict[str, Any]):
        data: Dict[str, Any] = raw_message["data"]
        try:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(data["symbol"])
        except KeyError:
            return
        if trading_pair not in self._trading_pairs:
            return

        self._last_mark_prices[trading_pair] = Decimal(data["mark_price"])
        self._last_index_prices[trading_pair] = Decimal(data["index_price"])

    async def _request_complete_funding_info(self, trading_pair: str):
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        data = await self._connector._api_get(
            path_url=CONSTANTS.FUNDING_INFO_URL,
            params={"symbol": ex_trading_pair})
        return data

    async def _request_complete_contract_details(self, trading_pair: str):
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        data = await self._connector._api_get(
            path_url=CONSTANTS.EXCHANGE_INFO_URL,
            params={"symbol": ex_trading_pair})
        return data

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        params = {
            "symbol": ex_trading_pair,
        }
        data = await self._connector._api_get(
            path_url=CONSTANTS.SNAPSHOT_REST_URL,
            params=params)
        return data

    @staticmethod
    def _parse_trade_way(way: int) -> str:
        way_to_trade_type = {
            1: TradeType.BUY,  # buy_open_long or sell_open_short (treated as buy)
            2: TradeType.BUY,  # buy_open_long or sell_close_long (treated as buy)
            3: TradeType.SELL,  # buy_close_short or sell_open_short (treated as sell)
            4: TradeType.SELL,  # buy_close_short or sell_close_long (treated as sell)
            5: TradeType.BUY,  # sell_open_short or buy_open_long (treated as buy)
            6: TradeType.SELL,  # sell_open_short or buy_close_short (treated as sell)
            7: TradeType.SELL,  # sell_close_long or buy_open_long (treated as sell)
            8: TradeType.BUY,  # sell_close_long or buy_close_short (treated as buy)
        }
        return way_to_trade_type.get(way)  # Default to "unknown" if way is invalid

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        """Get the next subscription ID and increment the counter."""
        subscribe_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return subscribe_id

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribe to order book channels for a single trading pair dynamically.

        :param trading_pair: The trading pair to subscribe to.
        :return: True if subscription was successful, False otherwise.
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot subscribe to {trading_pair}: WebSocket connection not established."
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            stream_id_channel_pairs = [
                CONSTANTS.ORDER_BOOK_CHANNEL,
                CONSTANTS.TRADE_STREAM_CHANNEL,
                CONSTANTS.FUNDING_INFO_CHANNEL,
                CONSTANTS.TICKERS_CHANNEL,
            ]

            for channel in stream_id_channel_pairs:
                params = [f"{channel}{f':{symbol.upper()}' if channel != 'futures/ticker' else ''}"]
                payload = {
                    "action": "subscribe",
                    "args": params,
                }
                subscribe_request: WSJSONRequest = WSJSONRequest(payload)
                await self._ws_assistant.send(subscribe_request)

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Successfully subscribed to {trading_pair}")
            return True

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(f"Error subscribing to {trading_pair}: {e}")
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribe from order book channels for a single trading pair dynamically.

        :param trading_pair: The trading pair to unsubscribe from.
        :return: True if unsubscription was successful, False otherwise.
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket connection not established."
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            stream_id_channel_pairs = [
                CONSTANTS.ORDER_BOOK_CHANNEL,
                CONSTANTS.TRADE_STREAM_CHANNEL,
                CONSTANTS.FUNDING_INFO_CHANNEL,
                CONSTANTS.TICKERS_CHANNEL,
            ]

            for channel in stream_id_channel_pairs:
                params = [f"{channel}{f':{symbol.upper()}' if channel != 'futures/ticker' else ''}"]
                payload = {
                    "action": "unsubscribe",
                    "args": params,
                }
                unsubscribe_request: WSJSONRequest = WSJSONRequest(payload)
                await self._ws_assistant.send(unsubscribe_request)

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Successfully unsubscribed from {trading_pair}")
            return True

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(f"Error unsubscribing from {trading_pair}: {e}")
            return False



from typing import Any, Dict, Optional

from hummingbot.connector.exchange.vertex.vertex_utils import convert_from_x18, convert_timestamp
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType


class VertexOrderBook(OrderBook):
    @classmethod
    def snapshot_message_from_exchange_websocket(
        cls, msg: Dict[str, Any], timestamp: float, metadata: Optional[Dict] = None
    ) -> OrderBookMessage:
        """
        Creates a snapshot message with the order book snapshot message
        :param msg: the response from the exchange when requesting the order book snapshot
        :param timestamp: the snapshot timestamp
        :param metadata: a dictionary with extra information to add to the snapshot data
        :return: a snapshot message with the snapshot information received from the exchange
        """
        if metadata:
            msg.update(metadata)
        ts = msg["data"]["timestamp"]
        return OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            {
                "trading_pair": msg["trading_pair"],
                "update_id": int(ts),
                "bids": convert_from_x18(msg["data"]["bids"]),
                "asks": convert_from_x18(msg["data"]["asks"]),
            },
            timestamp=timestamp,
        )

    @classmethod
    def snapshot_message_from_exchange_rest(
        cls, msg: Dict[str, Any], timestamp: float, metadata: Optional[Dict] = None
    ) -> OrderBookMessage:
        """
        Creates a snapshot message with the order book snapshot message
        :param msg: the response from the exchange when requesting the order book snapshot
        :param timestamp: the snapshot timestamp
        :param metadata: a dictionary with extra information to add to the snapshot data
        :return: a snapshot message with the snapshot information received from the exchange
        """
        if metadata:
            msg.update(metadata)
        ts = msg["data"]["timestamp"]
        return OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            {
                "trading_pair": msg["trading_pair"],
                "update_id": int(ts),
                "bids": convert_from_x18(msg["data"]["bids"]),
                "asks": convert_from_x18(msg["data"]["asks"]),
            },
            timestamp=timestamp,
        )

    @classmethod
    def diff_message_from_exchange(
        cls, msg: Dict[str, Any], timestamp: Optional[float] = None, metadata: Optional[Dict] = None
    ) -> OrderBookMessage:
        """
        Creates a diff message with the changes in the order book received from the exchange
        :param msg: the changes in the order book
        :param timestamp: the timestamp of the difference
        :param metadata: a dictionary with extra information to add to the difference data
        :return: a diff message with the changes in the order book notified by the exchange
        """
        if metadata:
            msg.update(metadata)
        ts = convert_timestamp(msg["last_max_timestamp"])
        return OrderBookMessage(
            OrderBookMessageType.DIFF,
            {
                "trading_pair": msg["trading_pair"],
                "update_id": int(msg["last_max_timestamp"]),
                "bids": convert_from_x18(msg["bids"]),
                "asks": convert_from_x18(msg["asks"]),
            },
            timestamp=ts,
        )

    @classmethod
    def trade_message_from_exchange(cls, msg: Dict[str, Any], metadata: Optional[Dict] = None):
        """
        Creates a trade message with the information from the trade event sent by the exchange
        :param msg: the trade event details sent by the exchange
        :param metadata: a dictionary with extra information to add to trade message
        :return: a trade message with the details of the trade as provided by the exchange
        """
        if metadata:
            msg.update(metadata)
        ts = convert_timestamp(msg["timestamp"])
        return OrderBookMessage(
            OrderBookMessageType.TRADE,
            {
                "trading_pair": msg["trading_pair"],
                "trade_type": float(TradeType.BUY.value) if msg["is_taker_buyer"] else float(TradeType.SELL.value),
                "trade_id": int(msg["timestamp"]),
                "update_id": int(msg["timestamp"]),
                "price": convert_from_x18(msg["price"]),
                "amount": convert_from_x18(msg["taker_qty"]),
            },
            timestamp=ts,
        )



import time
from typing import Optional

import hummingbot.connector.exchange.cube.cube_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler


async def get_current_server_time(
    throttler: Optional[AsyncThrottler] = None,
    domain: str = CONSTANTS.DEFAULT_DOMAIN,
) -> float:
    return time.time()



import asyncio
import logging
import time
from typing import Optional

import hummingbot.connector.exchange.bybit.bybit_constants as CONSTANTS
import hummingbot.connector.exchange.bybit.bybit_web_utils as web_utils
from hummingbot.connector.exchange.bybit.bybit_auth import BybitAuth
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger


class BybitAPIUserStreamDataSource(UserStreamTrackerDataSource):

    HEARTBEAT_TIME_INTERVAL = 30.0

    _bausds_logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: BybitAuth,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 api_factory: Optional[WebAssistantsFactory] = None,
                 throttler: Optional[AsyncThrottler] = None,
                 time_synchronizer: Optional[TimeSynchronizer] = None):
        super().__init__()
        self._auth: BybitAuth = auth
        self._time_synchronizer = time_synchronizer
        self._last_recv_time: float = 0
        self._domain = domain
        self._throttler = throttler
        self._api_factory = api_factory or web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)
        self._ws_assistant: Optional[WSAssistant] = None
        self._last_ws_message_sent_timestamp = 0

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._bausds_logger is None:
            cls._bausds_logger = logging.getLogger(__name__)
        return cls._bausds_logger

    @property
    def last_recv_time(self) -> float:
        """
        Returns the time of the last received message
        :return: the timestamp of the last received message in seconds
        """
        if self._ws_assistant:
            return self._ws_assistant.last_recv_time
        return 0

    async def listen_for_user_stream(self, output: asyncio.Queue):
        """
        Connects to the user private channel in the exchange using a websocket connection. With the established
        connection listens to all balance events and order updates provided by the exchange, and stores them in the
        output queue
        :param output: the queue to use to store the received messages
        """
        while True:
            try:
                ws: WSAssistant = await self._connected_websocket_assistant(self._domain)
                await self._subscribe_channels(ws)
                self._last_ws_message_sent_timestamp = self._time()
                while True:
                    try:
                        seconds_until_next_ping = (
                            CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL -
                            (self._time() - self._last_ws_message_sent_timestamp)
                        )
                        await asyncio.wait_for(
                            self._process_ws_messages(ws=ws, output=output), timeout=seconds_until_next_ping)
                    except asyncio.TimeoutError:
                        await self._ping_server(ws)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error while listening to user stream. Retrying after 5 seconds...")
            finally:
                # Make sure no background task is leaked.
                ws and await ws.disconnect()
                await self._sleep(5)

    async def _ping_server(self, ws: WSAssistant):
        ping_time = self._time()
        payload = {
            "op": "ping",
            "args": int(ping_time * 1e3)
        }
        ping_request = WSJSONRequest(payload=payload)
        await ws.send(request=ping_request)
        self._last_ws_message_sent_timestamp = ping_time

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            payload = {
                "op": "subscribe",
                "args": [f"{CONSTANTS.WS_SUBSCRIPTION_ORDERS_ENDPOINT_NAME}"],
            }
            subscribe_orders_request = WSJSONRequest(payload)
            payload = {
                "op": "subscribe",
                "args": [f"{CONSTANTS.WS_SUBSCRIPTION_EXECUTIONS_ENDPOINT_NAME}"],
            }
            subscribe_executions_request = WSJSONRequest(payload)
            payload = {
                "op": "subscribe",
                "args": [f"{CONSTANTS.WS_SUBSCRIPTION_WALLET_ENDPOINT_NAME}"],
            }
            subscribe_wallet_request = WSJSONRequest(payload)

            await ws.send(subscribe_orders_request)
            await ws.send(subscribe_executions_request)
            await ws.send(subscribe_wallet_request)

            self.logger().info("Subscribed to private orders, executions and wallet channels")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to private channels...",
                exc_info=True
            )
            raise

    async def _authenticate_connection(self, ws: WSAssistant):
        """
        Sends the authentication message.
        :param ws: the websocket assistant used to connect to the exchange
        """
        request: WSJSONRequest = WSJSONRequest(
            payload=self._auth.generate_ws_auth_message()
        )
        await ws.send(request)

    async def _process_ws_messages(self, ws: WSAssistant, output: asyncio.Queue):
        async for ws_response in ws.iter_messages():
            data = ws_response.data
            if "op" in data:
                if data.get("op") == "auth":
                    await self._process_ws_auth_msg(data)
                elif data.get("op") == "subscribe":
                    if data.get("success") is False:
                        self.logger().error(
                            "Unexpected error occurred subscribing to private channels...",
                            exc_info=True
                        )
                continue
            topic = data.get("topic")
            channel = ""
            if topic == CONSTANTS.WS_SUBSCRIPTION_ORDERS_ENDPOINT_NAME:
                channel = CONSTANTS.PRIVATE_ORDER_CHANNEL
            elif topic == CONSTANTS.WS_SUBSCRIPTION_EXECUTIONS_ENDPOINT_NAME:
                channel = CONSTANTS.PRIVATE_TRADE_CHANNEL
            elif topic == CONSTANTS.WS_SUBSCRIPTION_WALLET_ENDPOINT_NAME:
                channel = CONSTANTS.PRIVATE_WALLET_CHANNEL
            else:
                output.put_nowait(data)
            if channel:
                data["channel"] = channel
                output.put_nowait(data)

    async def _process_ws_auth_msg(self, data: dict):
        if not data.get("success"):
            raise IOError(f"Private channel authentication failed - {data['ret_msg']}")
        else:
            self.logger().info("Private channel authentication success.")

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    async def _connected_websocket_assistant(self, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> WSAssistant:
        ws: WSAssistant = await self._get_ws_assistant()
        await ws.connect(
            ws_url=CONSTANTS.WSS_PRIVATE_URL[domain],
            ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL
        )
        await self._authenticate_connection(ws)
        return ws

    def _get_server_timestamp(self):
        return web_utils.get_current_server_time()

    def _time(self):
        return time.time()



import base64
import datetime
import hashlib
import hmac
from collections import OrderedDict
from typing import Any, Dict
from urllib.parse import urlencode

from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSJSONRequest

HTX_HOST_NAME = "api.huobi.pro"


class HtxAuth(AuthBase):
    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key: str = api_key
        self.hostname: str = HTX_HOST_NAME
        self.secret_key: str = secret_key
        self.time_provider = time_provider

    @staticmethod
    def keysort(dictionary: Dict[str, str]) -> Dict[str, str]:
        return OrderedDict(sorted(dictionary.items(), key=lambda t: t[0]))

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:

        auth_params = self.generate_auth_params_for_REST(request=request)
        request.params = auth_params

        return request

    async def ws_authenticate(self, request: WSJSONRequest) -> WSJSONRequest:
        return request  # pass-through

    def generate_auth_params_for_REST(self, request: RESTRequest) -> Dict[str, Any]:
        timestamp = datetime.datetime.fromtimestamp(self.time_provider.time(), datetime.UTC).strftime("%Y-%m-%dT%H:%M:%S")
        path_url = f"/v1{request.url.split('v1')[-1]}"
        params = request.params or {}
        params.update({
            "AccessKeyId": self.api_key,
            "SignatureMethod": "HmacSHA256",
            "SignatureVersion": "2",
            "Timestamp": timestamp
        })
        sorted_params = self.keysort(params)
        signature = self.generate_signature(method=request.method.value.upper(),
                                            path_url=path_url,
                                            params=sorted_params,
                                            )
        sorted_params["Signature"] = signature
        return sorted_params

    def generate_auth_params_for_WS(self, request: WSJSONRequest) -> Dict[str, Any]:
        timestamp = datetime.datetime.fromtimestamp(self.time_provider.time(), datetime.UTC).strftime("%Y-%m-%dT%H:%M:%S")
        path_url = "/ws/v2"
        params = request.payload.get("params") or {}
        params.update({
            "accessKey": self.api_key,
            "signatureMethod": "HmacSHA256",
            "signatureVersion": "2.1",
            "timestamp": timestamp
        })
        sorted_params = self.keysort(params)
        signature = self.generate_signature(method="get",
                                            path_url=path_url,
                                            params=sorted_params,
                                            )
        sorted_params["signature"] = signature
        sorted_params["authType"] = "api"
        return sorted_params

    def generate_signature(self,
                           method: str,
                           path_url: str,
                           params: Dict[str, Any],
                           ) -> str:

        query_endpoint = path_url
        encoded_params_str = urlencode(params)
        payload = "\n".join([method.upper(), self.hostname, query_endpoint, encoded_params_str])
        digest = hmac.new(self.secret_key.encode("utf8"), payload.encode("utf8"), hashlib.sha256).digest()
        signature_b64 = base64.b64encode(digest).decode()

        return signature_b64



from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, Field, SecretStr

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = True
EXAMPLE_PAIR = "BTC-USDT"
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.001"),
    taker_percent_fee_decimal=Decimal("0.001"),
)


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair
    :return: True if the trading pair is enabled, False otherwise
    """
    return exchange_info.get("showStatus") is True


class BybitConfigMap(BaseConnectorConfigMap):
    connector: str = "bybit"
    bybit_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Bybit API key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    bybit_api_secret: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Bybit API secret",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="bybit")


KEYS = BybitConfigMap.model_construct()

OTHER_DOMAINS = ["bybit_testnet"]
OTHER_DOMAINS_PARAMETER = {"bybit_testnet": "bybit_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"bybit_testnet": "BTC-USDT"}
OTHER_DOMAINS_DEFAULT_FEES = {"bybit_testnet": DEFAULT_FEES}


class BybitTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = "bybit_testnet"
    bybit_testnet_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Bybit Testnet API Key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    bybit_testnet_api_secret: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Bybit Testnet API secret",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="bybit_testnet")


OTHER_DOMAINS_KEYS = {"bybit_testnet": BybitTestnetConfigMap.model_construct()}



import abc

from hummingbot.core.web_assistant.connections.data_types import WSRequest


class WSPreProcessorBase(abc.ABC):
    """An interface class that enables functionality injection into the `WSAssistant`.

    The logic provided by a class implementing this interface is applied to a request
    before it is sent out to the server.
    """

    @abc.abstractmethod
    async def pre_process(self, request: WSRequest) -> WSRequest:
        ...



from decimal import Decimal
from enum import Enum
from typing import Optional

from hummingbot.core.data_type.in_flight_order import InFlightOrder


class CloseType(Enum):
    TIME_LIMIT = 1
    STOP_LOSS = 2
    TAKE_PROFIT = 3
    EXPIRED = 4
    EARLY_STOP = 5
    TRAILING_STOP = 6
    INSUFFICIENT_BALANCE = 7
    FAILED = 8
    COMPLETED = 9
    POSITION_HOLD = 10


class TrackedOrder:
    def __init__(self, order_id: Optional[str] = None):
        self._order_id = order_id
        self._order = None

    @property
    def order_id(self):
        return self._order_id

    @order_id.setter
    def order_id(self, order_id: str):
        self._order_id = order_id

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, order: InFlightOrder):
        self._order = order

    @property
    def creation_timestamp(self):
        if self.order:
            return self.order.creation_timestamp
        else:
            return None

    @property
    def price(self):
        if self.order:
            return self.order.price
        else:
            return None

    @property
    def last_update_timestamp(self):
        if self.order:
            return self.order.last_update_timestamp
        else:
            return None

    @property
    def average_executed_price(self):
        if self.order:
            return self.order.average_executed_price or self.order.price
        else:
            return Decimal("0")

    @property
    def executed_amount_base(self):
        if self.order:
            return self.order.executed_amount_base
        else:
            return Decimal("0")

    @property
    def executed_amount_quote(self):
        if self.order:
            return self.order.executed_amount_quote
        else:
            return Decimal("0")

    @property
    def fee_asset(self):
        if self.order and len(self.order.order_fills) > 0:
            return list(self.order.order_fills.values())[0].fee_asset
        else:
            return None

    @property
    def cum_fees_base(self):
        if self.order:
            return self.order.cumulative_fee_paid(token=self.order.base_asset)
        else:
            return Decimal("0")

    @property
    def cum_fees_quote(self):
        if self.order:
            return self.order.cumulative_fee_paid(token=self.order.quote_asset)
        else:
            return Decimal("0")

    @property
    def is_done(self):
        if self.order:
            return self.order.is_done
        else:
            return False

    @property
    def is_open(self):
        if self.order:
            return self.order.is_open
        else:
            return False

    @property
    def is_filled(self):
        if self.order:
            return self.order.is_filled
        else:
            return False



# distutils: language=c++

from hummingbot.core.data_type.OrderExpirationEntry cimport OrderExpirationEntry as CPPOrderExpirationEntry


cdef class OrderExpirationEntry:
    cdef:
        CPPOrderExpirationEntry _cpp_order_expiration_entry


cdef OrderExpirationEntry c_create_order_expiration_from_cpp_order_expiration(const CPPOrderExpirationEntry cpp_order_expiration_entry)



import asyncio
import logging
import time
from collections import deque
from typing import Awaitable, Deque

import numpy

from hummingbot.logger import HummingbotLogger


class TimeSynchronizer:
    """
    Used to synchronize the local time with the server's time.
    This class is useful when timestamp-based signatures are required by the exchange for authentication.
    Upon receiving a timestamped message from the server, use `update_server_time_offset_with_time_provider`
    to synchronize local time with the server's time.
    """

    NaN = float("nan")
    _logger = None

    def __init__(self):
        self._time_offset_ms: Deque[float] = deque(maxlen=5)
        self._lock = asyncio.Lock()

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    @property
    def time_offset_ms(self) -> float:
        if not self._time_offset_ms:
            offset = (self._time() - self._current_seconds_counter()) * 1e3
        else:
            median = numpy.median(self._time_offset_ms)
            weighted_average = numpy.average(self._time_offset_ms, weights=range(1, len(self._time_offset_ms) * 2 + 1, 2))
            offset = numpy.mean([median, weighted_average])

        return offset

    def add_time_offset_ms_sample(self, offset: float):
        self._time_offset_ms.append(offset)

    def clear_time_offset_ms_samples(self):
        self._time_offset_ms.clear()

    def time(self) -> float:
        """
        Returns the current time in seconds calculated base on the deviation samples.
        :return: Calculated current time considering the registered deviations
        """
        return self._current_seconds_counter() + self.time_offset_ms * 1e-3

    async def update_server_time_offset_with_time_provider(self, time_provider: Awaitable):
        """
        Executes the time_provider passed as parameter to obtain the current time, and adds a new sample in the
        internal list.

        :param time_provider: Awaitable object that returns the current time
        """
        try:
            local_before_ms: float = self._current_seconds_counter() * 1e3
            server_time_ms: float = await time_provider
            local_after_ms: float = self._current_seconds_counter() * 1e3
            local_server_time_pre_image_ms: float = (local_before_ms + local_after_ms) / 2.0
            time_offset_ms: float = server_time_ms - local_server_time_pre_image_ms
            self.add_time_offset_ms_sample(time_offset_ms)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().network("Error getting server time.", exc_info=True,
                                  app_warning_msg="Could not refresh server time. Check network connection.")
            time_provider.close()

    async def update_server_time_if_not_initialized(self, time_provider: Awaitable):
        """
        Executes the time_provider passed as parameter to obtain the current time, and adds a new sample in the
        internal list, ONLY if the current instance has not been updated yet.

        :param time_provider: Awaitable object that returns the current time
        """
        async with self._lock:
            if not self._time_offset_ms:
                await self.update_server_time_offset_with_time_provider(time_provider)
            else:
                # This is done to avoid the warning message from asyncio framework saying a coroutine was not awaited
                time_provider.close()

    def _current_seconds_counter(self):
        return time.perf_counter()

    def _time(self):
        return time.time()



import asyncio
import logging
from typing import Any, Dict, Optional

import aiohttp

from hummingbot.core.network_base import NetworkBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_retry import async_retry
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class LogServerClient(NetworkBase):
    lsc_logger: Optional[HummingbotLogger] = None
    _lsc_shared_instance: "LogServerClient" = None

    @classmethod
    def get_instance(cls, log_server_url: str = "https://api.coinalpha.com/reporting-proxy-v2/") -> "LogServerClient":
        if cls._lsc_shared_instance is None:
            cls._lsc_shared_instance = LogServerClient(log_server_url=log_server_url)
        return cls._lsc_shared_instance

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls.lsc_logger is None:
            cls.lsc_logger = logging.getLogger(__name__)
        return cls.lsc_logger

    def __init__(self, log_server_url: str = "https://api.coinalpha.com/reporting-proxy-v2/"):
        super().__init__()
        self.queue: asyncio.Queue = asyncio.Queue()
        self.consume_queue_task: Optional[asyncio.Task] = None
        self.log_server_url: str = log_server_url

    def request(self, req):
        if not self.started:
            self.start()
        self.queue.put_nowait(req)

    @async_retry(retry_count=3, exception_types=[asyncio.TimeoutError, EnvironmentError], raise_exp=True)
    async def send_log(self, session: aiohttp.ClientSession, request_dict: Dict[str, Any]):
        async with session.request(request_dict["method"], request_dict["url"], **request_dict["request_obj"]) as resp:
            resp_text = await resp.text()
            self.logger().debug(f"Sent logs: {resp.status} {resp.url} {resp_text} ",
                                extra={"do_not_send": True})
            if resp.status != 200 and resp.status not in {404, 405, 400}:
                raise EnvironmentError("Failed sending logs to log server.")

    async def consume_queue(self, session):
        while True:
            try:
                req = await self.queue.get()
                self.logger().debug(f"Remote logging payload: {req}")
                await self.send_log(session, req)
            except asyncio.CancelledError:
                raise
            except aiohttp.ClientError:
                self.logger().network("Network error sending logs.", exc_info=True, extra={"do_not_send": True})
                return
            except Exception:
                self.logger().network("Unexpected error sending logs.", exc_info=True, extra={"do_not_send": True})
                return

    async def request_loop(self):
        while True:
            loop = asyncio.get_event_loop()
            try:
                async with aiohttp.ClientSession(loop=loop,
                                                 connector=aiohttp.TCPConnector(verify_ssl=False)) as session:
                    await self.consume_queue(session)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network("Unexpected error running logging task.",
                                      exc_info=True, extra={"do_not_send": True})
                await asyncio.sleep(5.0)

    async def start_network(self):
        self.consume_queue_task = safe_ensure_future(self.request_loop())

    async def stop_network(self):
        if self.consume_queue_task is not None:
            self.consume_queue_task.cancel()
            self.consume_queue_task = None

    async def check_network(self) -> NetworkStatus:
        try:
            loop = asyncio.get_event_loop()
            async with aiohttp.ClientSession(loop=loop,
                                             connector=aiohttp.TCPConnector(verify_ssl=False)) as session:
                async with session.get(self.log_server_url) as resp:
                    if resp.status != 200:
                        raise Exception("Log proxy server is down.")
        except asyncio.CancelledError:
            raise
        except Exception:
            return NetworkStatus.NOT_CONNECTED
        return NetworkStatus.CONNECTED



# distutils: language=c++

import asyncio
from enum import Enum
import logging
from typing import Optional

from hummingbot.core.clock cimport Clock
from hummingbot.logger import HummingbotLogger
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.time_iterator import TimeIterator

NaN = float("nan")
ni_logger = None


class NetworkStatus(Enum):
    STOPPED = 0
    NOT_CONNECTED = 1
    CONNECTED = 2


cdef class NetworkIterator(TimeIterator):
    @classmethod
    def logger(cls) -> HummingbotLogger:
        global ni_logger
        if ni_logger is None:
            ni_logger = logging.getLogger(__name__)
        return ni_logger

    def __init__(self):
        super().__init__()
        self._network_status = NetworkStatus.STOPPED
        self._last_connected_timestamp = NaN
        self._check_network_interval = 10.0
        self._check_network_timeout = 5.0
        self._network_error_wait_time = 60.0
        self._check_network_task = None

    @property
    def network_status(self) -> NetworkStatus:
        return self._network_status

    @property
    def last_connected_timestamp(self) -> float:
        return self._last_connected_timestamp

    @last_connected_timestamp.setter
    def last_connected_timestamp(self, value):
        self._last_connected_timestamp = value

    @property
    def check_network_task(self) -> Optional[asyncio.Task]:
        return self._check_network_task

    @property
    def check_network_interval(self) -> float:
        return self._check_network_interval

    @check_network_interval.setter
    def check_network_interval(self, double interval):
        self._check_network_interval = interval

    @property
    def network_error_wait_time(self) -> float:
        return self._network_error_wait_time

    @network_error_wait_time.setter
    def network_error_wait_time(self, double wait_time):
        self._network_error_wait_time = wait_time

    @property
    def check_network_timeout(self) -> float:
        return self._check_network_timeout

    @check_network_timeout.setter
    def check_network_timeout(self, double timeout):
        self._check_network_timeout = timeout

    async def start_network(self):
        pass

    async def stop_network(self):
        pass

    async def check_network(self) -> NetworkStatus:
        self.logger().warning("check_network() has not been implemented!")
        return NetworkStatus.NOT_CONNECTED

    async def _check_network_loop(self):
        while True:
            last_status = self._network_status
            try:
                new_status = await asyncio.wait_for(self.check_network(), timeout=self._check_network_timeout)
                if new_status != last_status:
                    self._network_status = new_status
                    if self._network_status is NetworkStatus.CONNECTED:
                        self.logger().info(f"Network status has changed to {new_status}. Starting networking...")
                        await self.start_network()
                    else:
                        self.logger().info(f"Network status has changed to {new_status}. Stopping networking...")
                        await self.stop_network()
                await asyncio.sleep(self._check_network_interval)
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                self.logger().debug("Check network call has timed out. Network status is not connected.")
                self._network_status = NetworkStatus.NOT_CONNECTED
                await asyncio.sleep(self._check_network_interval)
            except Exception as e:
                self.logger().error(f"Unexpected error while checking for network status: {e}", exc_info=True)
                self._network_status = NetworkStatus.NOT_CONNECTED
                await asyncio.sleep(self._network_error_wait_time)


    cdef c_start(self, Clock clock, double timestamp):
        TimeIterator.c_start(self, clock, timestamp)
        self._check_network_task = safe_ensure_future(self._check_network_loop())
        self._network_status = NetworkStatus.NOT_CONNECTED

    cdef c_stop(self, Clock clock):
        TimeIterator.c_stop(self, clock)
        if self._check_network_task is not None:
            self._check_network_task.cancel()
            self._check_network_task = None
        self._network_status = NetworkStatus.STOPPED
        safe_ensure_future(self.stop_network())

    def start(self, clock: Clock, timestamp: float):
        self.c_start(clock, timestamp)

    def stop(self, clock: Clock):
        self.c_stop(clock)



import logging
from typing import List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.okx_perpetual_candles import constants as CONSTANTS
from hummingbot.logger import HummingbotLogger


class OKXPerpetualCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)

    @property
    def name(self):
        return f"okx_perpetual_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return f"{trading_pair}-SWAP"

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return True

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return True

    def _get_rest_candles_params(self, start_time: Optional[int] = None, end_time: Optional[int] = None,
                                 limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST) -> dict:
        params = {
            "instId": self._ex_trading_pair,
            "bar": CONSTANTS.INTERVALS[self.interval]
        }
        if start_time:
            params["before"] = start_time * 1000
        params["after"] = end_time * 1000
        return params

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        return [
            [
                self.ensure_timestamp_in_seconds(row[0]), row[1], row[2], row[3], row[4], row[6], row[7], 0., 0., 0.
            ]
            for row in data["data"]
        ][::-1]

    def ws_subscription_payload(self):
        candle_args = [{"channel": f"candle{CONSTANTS.INTERVALS[self.interval]}", "instId": self._ex_trading_pair}]
        return {
            "op": "subscribe",
            "args": candle_args
        }

    def _parse_websocket_message(self, data: dict):
        candles_row_dict = {}
        if data is not None and "data" in data:  # data will be None when the websocket is disconnected
            candles = data["data"][0]
            candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(candles[0])
            candles_row_dict["open"] = candles[1]
            candles_row_dict["high"] = candles[2]
            candles_row_dict["low"] = candles[3]
            candles_row_dict["close"] = candles[4]
            candles_row_dict["volume"] = candles[6]
            candles_row_dict["quote_asset_volume"] = candles[7]
            candles_row_dict["n_trades"] = 0.
            candles_row_dict["taker_buy_base_volume"] = 0.
            candles_row_dict["taker_buy_quote_volume"] = 0.
            return candles_row_dict



import hashlib
import hmac
import threading
import time
from typing import Dict, Optional

from hummingbot.core.utils.tracking_nonce import get_tracking_nonce_low_res
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.connections_factory import ConnectionsFactory
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest

ONE_HOUR = 3600


class NdaxAuth(AuthBase):
    """
    Auth class required by NDAX API
    """

    _instance = None
    _lock = threading.Lock()  # To ensure thread safety during instance creation

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:  # Double-checked locking
                    cls._instance = super(NdaxAuth, cls).__new__(cls)
        return cls._instance

    def __init__(self, uid: str, api_key: str, secret_key: str, account_name: str):
        if not hasattr(self, "_initialized"):  # Prevent reinitialization
            if len(uid) > 0:
                self._uid: str = uid
                self._account_id = 0
                self._api_key: str = api_key
                self._secret_key: str = secret_key
                self._account_name: str = account_name
                self._token: Optional[str] = None
                self._token_expiration: int = 0
                self._initialized = True

    @property
    def token(self) -> str:
        return self._token

    @token.setter
    def token(self, token: str):
        self._token = token

    @property
    def uid(self) -> int:
        return int(self._uid)

    @uid.setter
    def uid(self, uid: str):
        self._uid = uid

    @property
    def account_id(self) -> int:
        return int(self._account_id)

    @property
    def account_name(self) -> str:
        return self._account_name

    def generate_nonce(self):
        return str(get_tracking_nonce_low_res())

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        headers = {}
        if self._token is None or time.time() > self._token_expiration:
            rest_connection = await ConnectionsFactory().get_rest_connection()
            request = RESTRequest(
                method=RESTMethod.POST,
                url="https://api.ndax.io:8443/AP/Authenticate",
                endpoint_url="",
                params={},
                data={},
                headers=self.header_for_authentication(),
            )
            authentication_req = await rest_connection.call(request)
            authentication = await authentication_req.json()
            if authentication.get("Authenticated", False) is True:
                self._token = authentication["SessionToken"]
                self._token_expiration = time.time() + ONE_HOUR - 10
                self._uid = authentication["User"]["UserId"]
                self._account_id = int(authentication["User"]["AccountId"])
            else:
                raise Exception("Could not authenticate REST connection with NDAX")

        headers.update({"APToken": self._token, "Content-Type": "application/json"})
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated.
        """
        return request

    def header_for_authentication(self) -> Dict[str, str]:
        """
        Generates authentication headers
        :return: a dictionary of auth headers
        """

        nonce = self.generate_nonce()
        raw_signature = nonce + str(self._uid) + self._api_key

        auth_info = {
            "Nonce": nonce,
            "APIKey": self._api_key,
            "Signature": hmac.new(
                self._secret_key.encode("utf-8"), raw_signature.encode("utf-8"), hashlib.sha256
            ).hexdigest(),
            "UserId": str(self._uid),
        }
        return auth_info



import asyncio
import math
from decimal import Decimal
from typing import Any, AsyncIterable, Dict, List, Optional, Tuple

from bidict import bidict
from dateutil.parser import parse as dateparse

import hummingbot.connector.exchange.btc_markets.btc_markets_constants as CONSTANTS
import hummingbot.connector.exchange.btc_markets.btc_markets_utils as utils
import hummingbot.connector.exchange.btc_markets.btc_markets_web_utils as web_utils
from hummingbot.connector.exchange.btc_markets.btc_markets_api_order_book_data_source import (
    BtcMarketsAPIOrderBookDataSource,
)
from hummingbot.connector.exchange.btc_markets.btc_markets_api_user_stream_data_source import (
    BtcMarketsAPIUserStreamDataSource,
)
from hummingbot.connector.exchange.btc_markets.btc_markets_auth import BtcMarketsAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

s_logger = None
s_decimal_0 = Decimal(0)
s_decimal_NaN = Decimal("nan")


class BtcMarketsExchange(ExchangePyBase):
    """
    BtcMarketsExchange connects with BtcMarkets exchange and provides order book pricing, user account tracking and
    trading functionality.
    """

    SHORT_POLL_INTERVAL = 5.0
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0
    LONG_POLL_INTERVAL = 120.0

    web_utils = web_utils

    def __init__(self,
                 btc_markets_api_key: str,
                 btc_markets_api_secret: str,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100"),
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 ):
        """
        :param btc_markets_api_key: The API key to connect to private BTCMarkets APIs.
        :param btc_markets_api_secret: The API secret.
        :param trading_pairs: The market trading pairs which to track order book data.
        :param trading_required: Whether actual trading is needed.
        """
        self._api_key: str = btc_markets_api_key
        self._secret_key: str = btc_markets_api_secret
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._domain = domain
        super().__init__(balance_asset_limit, rate_limits_share_pct)
        self.real_time_balance_update = False

    @property
    def authenticator(self):
        return BtcMarketsAuth(
            api_key=self._api_key,
            secret_key=self._secret_key,
            time_provider=self._time_synchronizer)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            auth=self.authenticator)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BtcMarketsAPIOrderBookDataSource(
            trading_pairs=self.trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BtcMarketsAPIUserStreamDataSource(
            auth=self.authenticator,
            trading_pairs=self.trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory)

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return CONSTANTS.DEFAULT_DOMAIN

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.MARKETS_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.MARKETS_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.SERVER_TIME_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def name_cap(self) -> str:
        return self.name.capitalize()

    @property
    def name(self) -> str:
        return CONSTANTS.EXCHANGE_NAME

    @staticmethod
    def btc_markets_order_type(order_type: OrderType) -> str:
        return order_type.name  # .upper()

    @staticmethod
    def to_hb_order_type(btc_markets_type: str) -> OrderType:
        return OrderType[btc_markets_type]

    # https://docs.btcmarkets.net/v3/#tag/OrderTypes
    def supported_order_types(self):
        return [OrderType.MARKET, OrderType.LIMIT, OrderType.LIMIT_MAKER]

    # https://docs.btcmarkets.net/v3/#tag/ErrorCodes
    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_code = str(request_exception)
        is_time_synchronizer_related = CONSTANTS.INVALID_TIME_WINDOW in error_code or CONSTANTS.INVALID_TIMESTAMP in error_code or CONSTANTS.INVALID_AUTH_TIMESTAMP in error_code or CONSTANTS.INVALID_AUTH_SIGNATURE in error_code
        return is_time_synchronizer_related

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return str(CONSTANTS.ORDER_NOT_FOUND) in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return str(CONSTANTS.ORDER_NOT_FOUND) in str(cancelation_exception)

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        response = await self._api_delete(
            path_url=f"{CONSTANTS.ORDERS_URL}/{tracked_order.exchange_order_id}",
            is_auth_required=True,
            limit_id=f"{CONSTANTS.ORDERS_URL}"
        )
        cancelled = True if response["clientOrderId"] == order_id else False

        return cancelled

    async def _place_order(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Decimal,
        **kwargs
    ) -> Tuple[str, float]:
        order_result = None
        amount_str = f"{amount:f}"
        price_str = f"{price:f}"
        type_str = 'Bid' if trade_type is TradeType.BUY else 'Ask'

        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        post_data = {
            "marketId": symbol,
            "side": type_str,
            "amount": amount_str,
            "selfTrade": "P",  # prevents self trading
            "clientOrderId": order_id,
            "timeInForce": CONSTANTS.TIME_IN_FORCE_GTC
        }

        if order_type == OrderType.MARKET:
            post_data["type"] = "Market"
        elif order_type == OrderType.LIMIT:
            post_data["type"] = "Limit"
            post_data["price"] = price_str
        elif order_type == OrderType.LIMIT_MAKER:
            post_data["type"] = "Limit"
            post_data["price"] = price_str
            post_data["postOnly"] = "true"

        order_result = await self._api_post(
            path_url = CONSTANTS.ORDERS_URL,
            data = post_data,
            is_auth_required = True
        )
        exchange_order_id = str(order_result["orderId"])

        return exchange_order_id, self.current_timestamp

    def _get_fee(
        self,
        base_currency: str,
        quote_currency: str,
        order_type: OrderType,
        order_side: TradeType,
        amount: Decimal,
        price: Decimal = s_decimal_NaN,
        is_maker: Optional[bool] = None
    ) -> AddedToCostTradeFee:
        """
        Calculates the estimated fee an order would pay based on the connector configuration
        :param base_currency: the order base currency
        :param quote_currency: the order quote currency
        :param order_type: the type of order (MARKET, LIMIT, LIMIT_MAKER)
        :param order_side: if the order is for buying or selling
        :param amount: the order amount
        :param price: the order price
        :return: the estimated fee for the order
        """
        """
        To get trading fee, this function is simplified by using fee override configuration. Most parameters to this
        function are ignore except order_type. Use OrderType.LIMIT_MAKER to specify you want trading fee for
        maker order.
        """
        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        trading_pair = combine_to_hb_trading_pair(base=base_currency, quote=quote_currency)
        if trading_pair in self._trading_fees:
            fees_data = self._trading_fees[trading_pair]
            fee_value = Decimal(fees_data["makerFeeRate"]) if is_maker else Decimal(fees_data["takerFeeRate"])
            fee = AddedToCostTradeFee(percent=fee_value)
        else:
            fee = build_trade_fee(
                self.name,
                is_maker,
                base_currency=base_currency,
                quote_currency=quote_currency,
                order_type=order_type,
                order_side=order_side,
                amount=amount,
                price=price,
            )
        return fee

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        resp = await self._api_get(
            path_url=CONSTANTS.FEES_URL,
            is_auth_required=True,
            limit_id=CONSTANTS.FEES_URL
        )
        fees_json = resp["feeByMarkets"]
        for fee_json in fees_json:
            trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=fee_json["marketId"])
            self._trading_fees[trading_pair] = fee_json

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []
        try:
            if order.exchange_order_id is not None:
                all_fills_response = await self._request_order_fills(order=order)
                updates = self._create_order_fill_updates(order=order, fill_update=all_fills_response)
                trade_updates.extend(updates)
        except asyncio.CancelledError:
            raise
        except Exception as ex:
            is_error_caused_by_unexistent_order = '"code":"OrderNotFound"' in str(ex)
            if not is_error_caused_by_unexistent_order:
                raise

        return trade_updates

    async def _request_order_fills(self, order: InFlightOrder) -> Dict[str, Any]:
        orderId = await order.get_exchange_order_id()
        return await self._api_get(
            path_url=CONSTANTS.TRADES_URL,
            params={
                "orderId": orderId
            },
            is_auth_required=True,
            limit_id=CONSTANTS.TRADES_URL
        )

    async def _request_order_update(self, order: InFlightOrder) -> Dict[str, Any]:
        return await self._get_order_update(order.exchange_order_id)

    async def _get_order_update(self, orderId: int) -> Dict[str, Any]:
        return await self._api_get(
            path_url=f"{CONSTANTS.ORDERS_URL}/{orderId}",
            is_auth_required=True,
            limit_id=CONSTANTS.ORDERS_URL
        )

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        updated_order_data = await self._request_order_update(order=tracked_order)

        order_update = self._create_order_update(order=tracked_order, order_update=updated_order_data)
        return order_update

    async def _format_trading_rules(self, exchange_info_dict: Dict[str, Any]) -> List[TradingRule]:
        """
        Example:
                [
                    {
                        "marketId": "BTC-AUD",
                        "baseAssetName": "BTC",
                        "quoteAssetName": "AUD",
                        "minOrderAmount": "0.0001",
                        "maxOrderAmount": "1000000",
                        "amountDecimals": "8",
                        "priceDecimals": "2",
                        "status": "Online"
                    },
                    {
                        "marketId": "LTC-AUD",
                        "baseAssetName": "LTC",
                        "quoteAssetName": "AUD",
                        "minOrderAmount": "0.001",
                        "maxOrderAmount": "1000000",
                        "amountDecimals": "8",
                        "priceDecimals": "2",
                        "status": "Post Only"
                    }
                ]
        """

        trading_pair_rules = exchange_info_dict
        retval = []
        for rule in trading_pair_rules:
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule["marketId"])

                min_order_size = Decimal(str(rule["minOrderAmount"]))
                # E.g. a price decimal of 2 means 0.01 incremental.
                price_decimal = Decimal(str(rule["priceDecimals"]))
                price_step = Decimal("1") / Decimal(str(math.pow(10, price_decimal)))
                amount_decimal = Decimal(str(rule["amountDecimals"]))
                amount_step = Decimal("1") / Decimal(str(math.pow(10, amount_decimal)))

                retval.append(
                    TradingRule(
                        trading_pair,
                        min_order_size=Decimal(min_order_size),
                        max_order_size=Decimal(str(rule["maxOrderAmount"])),
                        min_price_increment=Decimal(price_step),
                        min_base_amount_increment=Decimal(amount_step),
                    )
                )

            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {rule}. Skipping.")
        return retval

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are balance updates, order updates and trade events.
        """
        async for event_message in self._iter_user_event_queue():
            try:
                event_type = event_message.get("messageType")

                if event_type == CONSTANTS.HEARTBEAT:
                    continue
                elif event_type == CONSTANTS.ORDER_CHANGE_EVENT_TYPE:
                    exchange_order_id: Optional[str] = event_message.get("orderId")
                    client_order_id: Optional[str] = event_message.get("clientOrderId")
                    if client_order_id is None:
                        infligthOrder = await self._get_order_update(exchange_order_id)
                        client_order_id: Optional[str] = infligthOrder.get("clientOrderId")

                    fillable_order = self._order_tracker.all_fillable_orders.get(client_order_id)
                    updatable_order = self._order_tracker.all_updatable_orders.get(client_order_id)

                    new_state = CONSTANTS.ORDER_STATE[event_message["status"]]
                    event_timestamp = int(dateparse(event_message["timestamp"]).timestamp())

                    if fillable_order is not None:
                        is_fill_candidate_by_state = new_state in [OrderState.PARTIALLY_FILLED, OrderState.FILLED]
                        # is_fill_candidate_by_amount = fillable_order.executed_amount_base < Decimal(event_message["filled_size"])
                        if is_fill_candidate_by_state:  # and is_fill_candidate_by_amount:
                            try:
                                for trade in event_message["trades"]:
                                    fee = TradeFeeBase.new_spot_fee(
                                        fee_schema = self.trade_fee_schema(),
                                        trade_type = fillable_order.trade_type,
                                        percent_token = fillable_order.quote_asset,
                                        flat_fees = [TokenAmount(amount=Decimal(trade["fee"]), token = fillable_order.quote_asset)]
                                    )

                                    try:
                                        trade_update = TradeUpdate(
                                            trade_id=str(trade["tradeId"]),
                                            client_order_id=client_order_id,
                                            exchange_order_id=exchange_order_id,
                                            trading_pair=fillable_order.trading_pair,
                                            fee=fee,
                                            fill_base_amount=Decimal(trade["volume"]),
                                            fill_quote_amount=Decimal(trade["valueInQuoteAsset"]),
                                            fill_price=Decimal(trade["price"]),
                                            fill_timestamp=event_timestamp
                                        )

                                        self._order_tracker.process_trade_update(trade_update)
                                    except asyncio.CancelledError:
                                        raise
                                    except Exception:
                                        self.logger().exception(
                                            f"Unexpected error requesting order fills for {fillable_order.client_order_id}")

                            except asyncio.CancelledError:
                                raise
                            except Exception:
                                self.logger().exception(
                                    "Unexpected error requesting order fills for {fillable_order.client_order_id}")

                    if updatable_order is not None:
                        order_update = OrderUpdate(
                            trading_pair=updatable_order.trading_pair,
                            update_timestamp=event_timestamp,
                            new_state=new_state,
                            client_order_id=client_order_id,
                            exchange_order_id=exchange_order_id,
                        )
                        self._order_tracker.process_order_update(order_update=order_update)

                elif event_type == CONSTANTS.FUND_CHANGE_EVENT_TYPE:
                    asset_name = event_message.get("currency")
                    type = event_message.get("type")
                    status = event_message.get("status")
                    amount = Decimal(event_message.get("amount"))
                    if status == "Complete":
                        if type == "Deposit":
                            self._account_available_balances[asset_name] = self._account_available_balances[asset_name] + amount
                            self._account_balances[asset_name] = self._account_balances[asset_name] + amount
                        elif type == "Withdrawal":
                            self._account_balances[asset_name] = self._account_balances[asset_name] - amount
                    if status == "Pending Authorization" and type == "Withdrawal":
                        self._account_available_balances[asset_name] = self._account_available_balances[asset_name] - amount

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, any]]:
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Error while reading user events queue. Retrying after 1 second.")
                await asyncio.sleep(1.0)

    def _create_order_fill_updates(
        self,
        order: InFlightOrder,
        fill_update: Dict[str, Any]
    ) -> List[TradeUpdate]:
        updates = []
        fills_data = fill_update

        for fill_data in fills_data:
            fee = TradeFeeBase.new_spot_fee(
                fee_schema = self.trade_fee_schema(),
                trade_type = order.trade_type,
                percent_token = order.quote_asset,
                flat_fees = [TokenAmount(amount=Decimal(fill_data.get("fee")), token = order.quote_asset)]
            )

            trade_update = TradeUpdate(
                trade_id = str(fill_data.get("id")),
                client_order_id = fill_data.get("clientOrderId"),
                exchange_order_id = fill_data.get("orderId"),
                trading_pair = order.trading_pair,
                fee = fee,
                fill_base_amount = Decimal(fill_data.get("amount")),
                fill_price = Decimal(fill_data.get("price")),
                fill_quote_amount=Decimal(fill_data.get("amount")) * Decimal(fill_data["price"]),
                fill_timestamp = int(dateparse(fill_data.get("timestamp")).timestamp())
            )
            updates.append(trade_update)

        return updates

    def _create_order_update(self, order: InFlightOrder, order_update: Dict[str, Any]) -> OrderUpdate:
        new_state = CONSTANTS.ORDER_STATE[order_update["status"]]
        return OrderUpdate(
            trading_pair=order.trading_pair,
            update_timestamp = int(dateparse(order_update["creationTime"]).timestamp()),
            new_state = new_state,
            client_order_id = order.client_order_id,
            exchange_order_id = str(order_update["orderId"])
        )

    async def _get_balances(self):
        return await self._api_get(
            method=RESTMethod.GET,
            path_url=CONSTANTS.BALANCE_URL,
            is_auth_required=True,
            limit_id=CONSTANTS.BALANCE_URL
        )

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()
        account_info = await self._get_balances()

        for balance_entry in account_info:
            asset_name = balance_entry["assetName"]
            free_balance = Decimal(balance_entry["available"])
            total_balance = Decimal(balance_entry["balance"])
            self._account_available_balances[asset_name] = free_balance
            self._account_balances[asset_name] = total_balance
            remote_asset_names.add(asset_name)
        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    async def _sleep(self, delay: float):
        await asyncio.sleep(delay)

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List[Dict[str, Any]]):
        mapping = bidict()
        for symbol_data in filter(utils.is_exchange_information_valid, exchange_info):
            instrument_id = symbol_data["marketId"]
            trading_pair = combine_to_hb_trading_pair(
                base = symbol_data["baseAssetName"],
                quote = symbol_data["quoteAssetName"]
            )
            if instrument_id in mapping:
                self.logger().error(
                    f"Instrument ID {instrument_id} (trading pair {trading_pair}) already present in the map "
                    f"(with trading pair {mapping[instrument_id]})."
                )
                continue
            elif trading_pair in mapping.inverse:
                self.logger().error(
                    f"Trading pair {trading_pair} (instrument ID {instrument_id}) already present in the map "
                    f"(with ID {mapping.inverse[trading_pair]})."
                )
                continue
            mapping[instrument_id] = trading_pair

        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair)
        data = await self._api_request(
            method=RESTMethod.GET,
            path_url=f"{CONSTANTS.MARKETS_URL}/{trading_pair}/ticker",
            limit_id=CONSTANTS.MARKETS_URL
        )

        return float(data["lastPrice"])



import asyncio
import time
from dataclasses import dataclass, fields
from enum import Enum
from typing import Optional, Set

import pandas as pd
from bidict import bidict
from pandas import DataFrame

from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.network_base import NetworkBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant


class LiquidationSide(Enum):
    SHORT = "SHORT"  # Short position got liquidated (=> price went long)
    LONG = "LONG"  # Long position got liquidated (=> price went short)

    def __str__(self):
        return '%s' % self.value


@dataclass
class Liquidation:
    """
    Represents the information of a single liquidation
    """
    timestamp: int
    trading_pair: str
    quantity: float
    price: float
    side: LiquidationSide


class LiquidationsBase(NetworkBase):
    """
    This class serves as a base class for fetching and storing liquidation data from crypto exchanges. The storage
    is done in a time based manner - meaning an aggregation happens and you can decide how much history you wanto to keep.
    The class uses the WS Assistants for all the IO operations,
    """

    def __init__(self, trading_pairs: Set[str], max_retention_seconds: int):
        super().__init__()
        async_throttler = AsyncThrottler(rate_limits=self.rate_limits)
        self._api_factory = WebAssistantsFactory(throttler=async_throttler)
        self._max_retention_seconds = max_retention_seconds
        self._trading_pairs = trading_pairs
        self._liquidations = {}
        self._listen_liquidations_task: Optional[asyncio.Task] = None
        self._cleanup_task: Optional[asyncio.Task] = None
        self._subscribed_to_channels = False
        self._trading_pairs_map = bidict()

    async def start_network(self):
        """
        This method starts the network and starts a task for listen_for_subscriptions.
        """
        await self.stop_network()
        await self._fetch_and_map_trading_pairs()
        self._listen_liquidations_task = safe_ensure_future(self.listen_for_subscriptions())
        self._cleanup_task = safe_ensure_future(self._cleanup_old_liquidations_loop())
        self.logger().info("Liquidations feed ({}) started, keeping the last {}s of data".format(self.name,
                                                                                                 self._max_retention_seconds))
        self._subscribed_to_channels = True

    async def stop_network(self):
        """
        This method stops the network by canceling the _listen_liquidations_task and _cleanup_task.
        """
        if self._listen_liquidations_task is not None:
            self._listen_liquidations_task.cancel()
            self._listen_liquidations_task = None

        if self._cleanup_task is not None:
            self._cleanup_task.cancel()
            self._cleanup_task = None

    @property
    def ready(self):
        """
        Turns True if websockets are subscribed to the liquidations feed and network is basically started / up
        """
        return self._subscribed_to_channels

    @property
    def name(self):
        raise NotImplementedError

    @property
    def rest_url(self):
        raise NotImplementedError

    @property
    def health_check_url(self):
        raise NotImplementedError

    @property
    def wss_url(self):
        raise NotImplementedError

    @property
    def rate_limits(self):
        raise NotImplementedError

    async def check_network(self) -> NetworkStatus:
        raise NotImplementedError

    async def _cleanup_old_liquidations_loop(self):
        while True:
            self._cleanup_old_liquidations()
            await self._sleep(1.0)

    def _cleanup_old_liquidations(self):
        try:
            current_time_ms = int(time.time() * 1000)
            if self._liquidations:
                for trading_pair, liquidations in list(self._liquidations.items()):
                    self._liquidations[trading_pair] = [
                        liq for liq in liquidations if
                        current_time_ms - liq.timestamp < self._max_retention_seconds * 1000
                    ]
        except Exception:
            self.logger().exception(
                "Unexpected error occurred when cleaning up outdated liquidations. Retrying in 1 seconds...",
            )

    def liquidations_df(self, trading_pair=None) -> DataFrame:
        """
        This method returns the liquidations stored as a Pandas DataFrame.
        If no trading_pair is specified, all liquidations are returned in a single DataFrame.
        If the specified trading_pair has no data, an empty DataFrame is returned.
        """
        # Dynamically retrieve column names from the Liquidation dataclass
        column_names = [f.name for f in fields(Liquidation)]

        # Check if a specific trading pair is requested
        if trading_pair:
            # Retrieve the list of Liquidation objects for the specified trading pair
            liquidations = self._liquidations.get(trading_pair)
            # If no liquidations are found for the trading pair, return an empty DataFrame
            if not liquidations:
                return pd.DataFrame(columns=column_names)
            # Convert the list of dataclass instances to a DataFrame
            return pd.DataFrame([liq.__dict__ for liq in liquidations])
        else:
            # No specific trading pair is requested, combine all pairs
            all_liquidations = []
            # Iterate over all trading pairs and collect their liquidations
            for pair_liquidations in self._liquidations.values():
                all_liquidations.extend(pair_liquidations)
            # If no liquidations are collected, return an empty DataFrame
            if not all_liquidations:
                return pd.DataFrame(columns=column_names)
            # Convert the list of all liquidations to a DataFrame
            return pd.DataFrame([liq.__dict__ for liq in all_liquidations])

    def get_exchange_trading_pair(self, trading_pair):
        raise NotImplementedError

    async def listen_for_subscriptions(self):
        """
        Connects to the liquidations (=forceOrder) websocket endpoint and listens to the messages sent by the
        exchange.
        """
        ws: Optional[WSAssistant] = None
        while True:
            try:
                ws: WSAssistant = await self._connected_websocket_assistant()
                await self._subscribe_channels(ws)
                await self._process_websocket_messages(websocket_assistant=ws)
            except asyncio.CancelledError:
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(f"The websocket connection was closed ({connection_exception})")
            except Exception:
                self.logger().exception(
                    "Unexpected error occurred when listening to public liquidations. Retrying in 1 seconds...",
                )
                await self._sleep(1.0)
            finally:
                await self._on_order_stream_interruption(websocket_assistant=ws)

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=self.wss_url,
                         ping_timeout=30)
        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the liquidation events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        raise NotImplementedError

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        raise NotImplementedError

    async def _fetch_and_map_trading_pairs(self):
        """
        Used to fetch the trading pairs from the exchange and create the bidict mapping for hummingbot which is
        later used in _process_websocket_messages to be able to give the right reverse mapping from the exchange native
        pair (e.g. "BTC-USDT" is "BTCUSDT" on binance)
        """
        raise NotImplementedError

    async def _sleep(self, delay):
        """
        Function added only to facilitate patching the sleep in unit tests without affecting the asyncio module
        """
        await asyncio.sleep(delay)

    async def _on_order_stream_interruption(self, websocket_assistant: Optional[WSAssistant] = None):
        websocket_assistant and await websocket_assistant.disconnect()



import asyncio
import threading
from typing import TYPE_CHECKING

from hummingbot.client.config.client_config_map import AutofillImportEnum
from hummingbot.client.config.config_helpers import (
    format_config_file_name,
    load_strategy_config_map_from_file,
    short_strategy_name,
    validate_strategy_file,
)
from hummingbot.client.settings import CONF_PREFIX, STRATEGIES_CONF_DIR_PATH, required_exchanges
from hummingbot.core.utils.async_utils import safe_ensure_future

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class ImportCommand:

    def import_command(self,  # type: HummingbotApplication
                       file_name):
        if file_name is not None:
            file_name = format_config_file_name(file_name)

        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.import_command, file_name)
            return
        safe_ensure_future(self.import_config_file(file_name))

    async def import_config_file(self,  # type: HummingbotApplication
                                 file_name):
        self.app.clear_input()
        self.placeholder_mode = True
        self.app.hide_input = True
        required_exchanges.clear()
        if file_name is None:
            file_name = await self.prompt_a_file_name()
        if self.app.to_stop_config:
            self.app.to_stop_config = False
            return
        strategy_path = STRATEGIES_CONF_DIR_PATH / file_name
        try:
            config_map = await load_strategy_config_map_from_file(strategy_path)
        except Exception as e:
            self.notify(f'Strategy import error: {str(e)}')
            # Reset prompt settings
            self.placeholder_mode = False
            self.app.hide_input = False
            self.app.change_prompt(prompt=">>> ")
            raise
        self.strategy_file_name = file_name
        self.trading_core.strategy_name = (
            config_map.strategy
            if not isinstance(config_map, dict)
            else config_map.get("strategy").value  # legacy
        )
        self.strategy_config_map = config_map
        self.notify(f"Configuration from {self.strategy_file_name} file is imported.")
        self.placeholder_mode = False
        self.app.hide_input = False
        self.app.change_prompt(prompt=">>> ")
        try:
            all_status_go = await self.status_check_all()
        except asyncio.TimeoutError:
            self.strategy_file_name = None
            self.trading_core.strategy_name = None
            self.strategy_config_map = None
            raise
        if all_status_go:
            self.notify("\nEnter \"start\" to start market making.")
            autofill_import = self.client_config_map.autofill_import
            if autofill_import != AutofillImportEnum.disabled:
                self.app.set_text(autofill_import)

    async def prompt_a_file_name(self  # type: HummingbotApplication
                                 ):
        example = f"{CONF_PREFIX}{short_strategy_name('pure_market_making')}_{1}.yml"
        file_name = await self.app.prompt(prompt=f'Enter path to your strategy file (e.g. "{example}") >>> ')
        if self.app.to_stop_config:
            return
        file_path = STRATEGIES_CONF_DIR_PATH / file_name
        err_msg = validate_strategy_file(file_path)
        if err_msg is not None:
            self.notify(f"Error: {err_msg}")
            return await self.prompt_a_file_name()
        else:
            return file_name



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.exchange.okx import okx_constants as CONSTANTS, okx_utils, okx_web_utils as web_utils
from hummingbot.connector.exchange.okx.okx_api_order_book_data_source import OkxAPIOrderBookDataSource
from hummingbot.connector.exchange.okx.okx_api_user_stream_data_source import OkxAPIUserStreamDataSource
from hummingbot.connector.exchange.okx.okx_auth import OkxAuth
from hummingbot.connector.exchange_base import s_decimal_NaN
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class OkxExchange(ExchangePyBase):

    web_utils = web_utils

    def __init__(self,
                 okx_api_key: str,
                 okx_secret_key: str,
                 okx_passphrase: str,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100"),
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 okx_registration_sub_domain: str = "www"):
        """
        :param okx_registration_sub_domain: The subdomain to use - options are "www" (default), "app" (US users), or "my" (EEA users)
                              See: https://github.com/ccxt/ccxt/issues/24601
        """
        self.okx_api_key = okx_api_key
        self.okx_secret_key = okx_secret_key
        self.okx_passphrase = okx_passphrase
        self.okx_registration_sub_domain = okx_registration_sub_domain or "www"
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def authenticator(self):
        return OkxAuth(
            api_key=self.okx_api_key,
            secret_key=self.okx_secret_key,
            passphrase=self.okx_passphrase,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        return "okx"

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return CONSTANTS.get_okx_base_url(self.okx_registration_sub_domain)

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.CLIENT_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.OKX_INSTRUMENTS_PATH

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.OKX_INSTRUMENTS_PATH

    @property
    def check_network_request_path(self):
        return CONSTANTS.OKX_SERVER_TIME_PATH

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_description = str(request_exception)
        is_time_synchronizer_related = '"code":"50113"' in error_description
        return is_time_synchronizer_related

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        # _place_cancel already takes care of all expected exceptions.
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            auth=self._auth,
            domain=self.domain)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return OkxAPIOrderBookDataSource(
            trading_pairs=self.trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return OkxAPIUserStreamDataSource(
            auth=self._auth,
            connector=self,
            api_factory=self._web_assistants_factory)

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:

        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        fee = build_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )
        return fee

    async def _initialize_trading_pair_symbol_map(self):
        # This has to be reimplemented because the request requires an extra parameter
        try:
            exchange_info = await self._api_get(
                path_url=self.trading_pairs_request_path,
                params={"instType": "SPOT"},
            )
            self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)
        except Exception:
            self.logger().exception("There was an error requesting exchange info.")

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(okx_utils.is_exchange_information_valid, exchange_info["data"]):
            mapping[symbol_data["instId"]] = combine_to_hb_trading_pair(base=symbol_data["baseCcy"],
                                                                        quote=symbol_data["quoteCcy"])
        self._set_trading_pair_symbol_map(mapping)

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:

        data = {
            "clOrdId": order_id,
            "tdMode": "cash",
            "ordType": CONSTANTS.ORDER_TYPE_MAP[order_type],
            "side": trade_type.name.lower(),
            "instId": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
            "sz": str(amount),
        }
        if order_type.is_limit_type():
            data["px"] = f"{price:f}"
        else:
            # Specify that the order quantity for market orders is denominated in base currency
            data["tgtCcy"] = "base_ccy"

        exchange_order_id = await self._api_request(
            path_url=CONSTANTS.OKX_PLACE_ORDER_PATH,
            method=RESTMethod.POST,
            data=data,
            is_auth_required=True,
            limit_id=CONSTANTS.OKX_PLACE_ORDER_PATH,
        )
        data = exchange_order_id["data"][0]
        if data["sCode"] != "0":
            raise IOError(f"Error submitting order {order_id}: {data['sMsg']}")
        return str(data["ordId"]), self.current_timestamp

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        """
        This implementation specific function is called by _cancel, and returns True if successful
        """
        params = {
            "clOrdId": order_id,
            "instId": tracked_order.trading_pair
        }
        cancel_result = await self._api_post(
            path_url=CONSTANTS.OKX_ORDER_CANCEL_PATH,
            data=params,
            is_auth_required=True,
        )
        if cancel_result["data"][0]["sCode"] == "0":
            final_result = True
        elif cancel_result["data"][0]["sCode"] == "51400":
            # Cancelation failed because the order does not exist
            final_result = True
        elif cancel_result["data"][0]["sCode"] == "51401":
            # Cancelation failed because order has been cancelled
            final_result = True
        else:
            raise IOError(f"Error cancelling order {order_id}: {cancel_result}")

        return final_result

    async def get_last_traded_prices(self, trading_pairs: List[str] = None) -> Dict[str, float]:
        params = {"instType": "SPOT"}

        if trading_pairs and len(trading_pairs) == 1:
            return {trading_pairs[0]: await self._get_last_traded_price(trading_pairs[0])}
        resp_json = await self._api_get(
            path_url=CONSTANTS.OKX_TICKERS_PATH,
            params=params,
        )
        # some markets have no last traded price, so filter them out
        last_traded_prices = {ticker["instId"]: float(ticker["last"]) for ticker in resp_json["data"] if ticker["last"]}
        return last_traded_prices

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {"instId": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)}

        resp_json = await self._api_request(
            path_url=CONSTANTS.OKX_TICKER_PATH,
            params=params,
        )

        ticker_data, *_ = resp_json["data"]
        return float(ticker_data["last"])

    async def _update_balances(self):
        msg = await self._api_request(
            path_url=CONSTANTS.OKX_BALANCE_PATH,
            is_auth_required=True)

        if msg['code'] == '0':
            balances = msg['data'][0]['details']
        else:
            raise Exception(msg['msg'])

        self._account_available_balances.clear()
        self._account_balances.clear()

        for balance in balances:
            self._update_balance_from_details(balance_details=balance)

    def _update_balance_from_details(self, balance_details: Dict[str, Any]):
        equity_text = balance_details["eq"]
        available_equity_text = balance_details["availEq"]

        if equity_text and available_equity_text:
            total = Decimal(equity_text)
            available = Decimal(available_equity_text)
        else:
            available = Decimal(balance_details["availBal"])
            total = available + Decimal(balance_details["frozenBal"])
        self._account_balances[balance_details["ccy"]] = total
        self._account_available_balances[balance_details["ccy"]] = available

    async def _update_trading_rules(self):
        # This has to be reimplemented because the request requires an extra parameter
        # TODO: Normalize the rest requests so they can be used standalone
        exchange_info = await self._api_get(
            path_url=self.trading_rules_request_path,
            params={"instType": "SPOT"},
        )
        trading_rules_list = await self._format_trading_rules(exchange_info)
        self._trading_rules.clear()
        for trading_rule in trading_rules_list:
            self._trading_rules[trading_rule.trading_pair] = trading_rule
        self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)

    async def _format_trading_rules(self, raw_trading_pair_info: List[Dict[str, Any]]) -> List[TradingRule]:
        trading_rules = []

        for info in raw_trading_pair_info.get("data", []):
            try:
                if okx_utils.is_exchange_information_valid(exchange_info=info):
                    trading_rules.append(
                        TradingRule(
                            trading_pair=await self.trading_pair_associated_to_exchange_symbol(symbol=info["instId"]),
                            min_order_size=Decimal(info["minSz"]),
                            min_price_increment=Decimal(info["tickSz"]),
                            min_base_amount_increment=Decimal(info["lotSz"]),
                        )
                    )
            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {info}. Skipping.")
        return trading_rules

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _request_order_update(self, order: InFlightOrder) -> Dict[str, Any]:
        return await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.OKX_ORDER_DETAILS_PATH,
            params={
                "instId": await self.exchange_symbol_associated_to_pair(order.trading_pair),
                "clOrdId": order.client_order_id},
            is_auth_required=True)

    async def _request_order_fills(self, order: InFlightOrder) -> Dict[str, Any]:
        return await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.OKX_TRADE_FILLS_PATH,
            params={
                "instType": "SPOT",
                "instId": await self.exchange_symbol_associated_to_pair(order.trading_pair),
                "ordId": await order.get_exchange_order_id()},
            is_auth_required=True)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            all_fills_response = await self._request_order_fills(order=order)
            fills_data = all_fills_response["data"]

            for fill_data in fills_data:
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=order.trade_type,
                    percent_token=fill_data["feeCcy"],
                    flat_fees=[TokenAmount(amount=-Decimal(fill_data["fee"]), token=fill_data["feeCcy"])]
                )
                trade_update = TradeUpdate(
                    trade_id=str(fill_data["tradeId"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=str(fill_data["ordId"]),
                    trading_pair=order.trading_pair,
                    fee=fee,
                    fill_base_amount=Decimal(fill_data["fillSz"]),
                    fill_quote_amount=Decimal(fill_data["fillSz"]) * Decimal(fill_data["fillPx"]),
                    fill_price=Decimal(fill_data["fillPx"]),
                    fill_timestamp=int(fill_data["ts"]) * 1e-3,
                )
                trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        updated_order_data = await self._request_order_update(order=tracked_order)

        order_data = updated_order_data["data"][0]
        new_state = CONSTANTS.ORDER_STATE[order_data["state"]]

        order_update = OrderUpdate(
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=str(order_data["ordId"]),
            trading_pair=tracked_order.trading_pair,
            update_timestamp=int(order_data["uTime"]) * 1e-3,
            new_state=new_state,
        )
        return order_update

    async def _user_stream_event_listener(self):
        async for stream_message in self._iter_user_event_queue():
            try:
                args = stream_message.get("arg", {})
                channel = args.get("channel", None)

                if channel == CONSTANTS.OKX_WS_ORDERS_CHANNEL:
                    for data in stream_message.get("data", []):
                        order_status = CONSTANTS.ORDER_STATE[data["state"]]
                        client_order_id = data["clOrdId"]
                        trade_id = data["tradeId"]
                        fillable_order = self._order_tracker.all_fillable_orders.get(client_order_id)
                        updatable_order = self._order_tracker.all_updatable_orders.get(client_order_id)

                        if (fillable_order is not None
                                and order_status in [OrderState.PARTIALLY_FILLED, OrderState.FILLED]
                                and trade_id):
                            fee = TradeFeeBase.new_spot_fee(
                                fee_schema=self.trade_fee_schema(),
                                trade_type=fillable_order.trade_type,
                                percent_token=data["fillFeeCcy"],
                                flat_fees=[TokenAmount(amount=-Decimal(data["fillFee"]), token=data["fillFeeCcy"])]
                            )
                            trade_update = TradeUpdate(
                                trade_id=str(trade_id),
                                client_order_id=fillable_order.client_order_id,
                                exchange_order_id=str(data["ordId"]),
                                trading_pair=fillable_order.trading_pair,
                                fee=fee,
                                fill_base_amount=Decimal(data["fillSz"]),
                                fill_quote_amount=Decimal(data["fillSz"]) * Decimal(data["fillPx"]),
                                fill_price=Decimal(data["fillPx"]),
                                fill_timestamp=int(data["uTime"]) * 1e-3,
                            )
                            self._order_tracker.process_trade_update(trade_update)

                        if updatable_order is not None:
                            order_update = OrderUpdate(
                                trading_pair=updatable_order.trading_pair,
                                update_timestamp=int(data["uTime"]) * 1e-3,
                                new_state=order_status,
                                client_order_id=updatable_order.client_order_id,
                                exchange_order_id=str(data["ordId"]),
                            )
                            self._order_tracker.process_order_update(order_update=order_update)

                elif channel == CONSTANTS.OKX_WS_ACCOUNT_CHANNEL:
                    for data in stream_message.get("data", []):
                        for details in data.get("details", []):
                            self._update_balance_from_details(balance_details=details)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")
                await self._sleep(5.0)



from decimal import Decimal
from typing import List, Optional
import warnings

from hummingbot.client.config.trade_fee_schema_loader import TradeFeeSchemaLoader
from hummingbot.core.data_type.trade_fee import (
    TradeFeeBase,
    TokenAmount,
    TradeFeeSchema
)
from hummingbot.core.data_type.common import OrderType, PositionAction, TradeType


def build_trade_fee(
    exchange: str,
    is_maker: bool,
    base_currency: str,
    quote_currency: str,
    order_type: OrderType,
    order_side: TradeType,
    amount: Decimal,
    price: Decimal = Decimal("NaN"),
    extra_flat_fees: Optional[List[TokenAmount]] = None,
) -> TradeFeeBase:
    """
    WARNING: Do not use this method for order sizing. Use the `BudgetChecker` instead.

    Uses the exchange's `TradeFeeSchema` to build a `TradeFee`, given the trade parameters.
    """
    trade_fee_schema: TradeFeeSchema = TradeFeeSchemaLoader.configured_schema_for_exchange(exchange_name=exchange)
    fee_percent: Decimal = (
        trade_fee_schema.maker_percent_fee_decimal
        if is_maker
        else trade_fee_schema.taker_percent_fee_decimal
    )
    fixed_fees: List[TokenAmount] = (
        trade_fee_schema.maker_fixed_fees
        if is_maker
        else trade_fee_schema.taker_fixed_fees
    ).copy()
    if extra_flat_fees is not None and len(extra_flat_fees) > 0:
        fixed_fees = fixed_fees + extra_flat_fees
    trade_fee: TradeFeeBase = TradeFeeBase.new_spot_fee(
        fee_schema=trade_fee_schema,
        trade_type=order_side,
        percent=fee_percent,
        percent_token=trade_fee_schema.percent_fee_token,
        flat_fees=fixed_fees
    )
    return trade_fee


def build_perpetual_trade_fee(
    exchange: str,
    is_maker: bool,
    position_action: PositionAction,
    base_currency: str,
    quote_currency: str,
    order_type: OrderType,
    order_side: TradeType,
    amount: Decimal,
    price: Decimal = Decimal("NaN"),
) -> TradeFeeBase:
    """
    WARNING: Do not use this method for order sizing. Use the `BudgetChecker` instead.

    Uses the exchange's `TradeFeeSchema` to build a `TradeFee`, given the trade parameters.
    """
    trade_fee_schema = TradeFeeSchemaLoader.configured_schema_for_exchange(exchange_name=exchange)
    percent = trade_fee_schema.maker_percent_fee_decimal if is_maker else trade_fee_schema.taker_percent_fee_decimal
    fixed_fees = trade_fee_schema.maker_fixed_fees if is_maker else trade_fee_schema.taker_fixed_fees
    trade_fee = TradeFeeBase.new_perpetual_fee(
        fee_schema=trade_fee_schema,
        position_action=position_action,
        percent=percent,
        percent_token=trade_fee_schema.percent_fee_token,
        flat_fees=fixed_fees)
    return trade_fee


def estimate_fee(exchange: str, is_maker: bool) -> TradeFeeBase:
    """
    WARNING: This method is deprecated and remains only for backward compatibility.
    Use `build_trade_fee` and `build_perpetual_trade_fee` instead.

    Estimate the fee of a transaction on any blockchain.
    exchange is the name of the exchange to query.
    is_maker if true look at fee from maker side, otherwise from taker side.
    """
    warnings.warn(
        "The 'estimate_fee' method is deprecated, use 'build_trade_fee' and 'build_perpetual_trade_fee' instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    trade_fee = build_trade_fee(
        exchange,
        is_maker,
        base_currency="",
        quote_currency="",
        order_type=OrderType.LIMIT,
        order_side=TradeType.BUY,
        amount=Decimal("0"),
        price=Decimal("0"),
    )
    return trade_fee



import json
from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict, List

from eth_account.messages import encode_defunct
from web3 import Web3

from hummingbot.connector.derivative.derive_perpetual import (
    derive_perpetual_constants as CONSTANTS,
    derive_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.derive_perpetual.derive_perpetual_web_utils import MAX_INT_32, get_action_nonce
from hummingbot.connector.other.derive_common_utils import SignedAction, TradeModuleData
from hummingbot.connector.utils import to_0x_hex
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class DerivePerpetualAuth(AuthBase):
    """
    Auth class required by DerivePerpetual API
    """

    def __init__(self, api_key: str, api_secret: str, sub_id: int, trading_required: bool, domain: str):
        self._api_key: str = api_key
        self._api_secret: str = api_secret
        self._sub_id: int = sub_id
        self._trading_required: bool = trading_required
        self._w3 = Web3()
        self._domain = domain
        if trading_required:
            self.session_key_wallet = Web3().eth.account.from_key(self._api_secret)

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. OKX does not use this
        functionality
        """
        return request  # pass-through

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        if request.method == RESTMethod.POST:
            request.data = self.add_auth_to_params_post(params=json.loads(request.data), request=request)
        else:
            request.params = self.add_auth_to_params_post(params=request.params, request=request)

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(self.header_for_authentication())
        request.headers = headers

        return request

    def get_ws_auth_payload(self) -> List[Dict[str, Any]]:
        payload = {}
        timestamp = str(self.utc_now_ms())
        signature = to_0x_hex(self._w3.eth.account.sign_message(
            encode_defunct(text=timestamp), private_key=self._api_secret
        ).signature)
        """
        This method is intended to configure a websocket request to be authenticated. Dexalot does not use this
        functionality
        """
        payload["accept"] = 'application/json'
        payload["wallet"] = self._api_key
        payload["timestamp"] = timestamp
        payload["signature"] = signature
        return payload

    def add_auth_to_params_post(self, params: Dict[str, str], request):
        payload = {}
        data = params if params is not None else {}

        request_params = data

        if "type" in request_params:
            request_type = request_params.get("type")
            request_params.pop("type")
            if request_type == "order":
                action = self.sign(request_params)
            request_params = web_utils.order_to_call(request_params)
            request_params.update(action)
            payload.update(request_params)
        else:
            payload.update(request_params)

        return json.dumps(payload) if request.method == RESTMethod.POST else payload

    def sign(self, params):
        domain_seperator = CONSTANTS.DOMAIN_SEPARATOR if "testnet" not in self._domain else CONSTANTS.TESTNET_DOMAIN_SEPARATOR
        action_typehash = CONSTANTS.ACTION_TYPEHASH if "testnet" not in self._domain else CONSTANTS.TESTNET_ACTION_TYPEHASH
        action = SignedAction(
            subaccount_id=int(self._sub_id),
            owner=self._api_key,
            signer=self.session_key_wallet.address,
            signature_expiry_sec=MAX_INT_32,
            nonce=get_action_nonce(),
            module_address=CONSTANTS.TRADE_MODULE_ADDRESS,
            module_data=TradeModuleData(
                asset_address=params["asset_address"],
                sub_id=int(params["sub_id"]),
                limit_price=(Decimal(params["limit_price"])),
                amount=Decimal(params["amount"]),
                max_fee=Decimal(params["max_fee"]),
                recipient_id=int(params["recipient_id"]),
                is_bid=params["is_bid"],
            ),
            DOMAIN_SEPARATOR=domain_seperator,  # from Protocol Constants table in docs.derive_perpetual.xyz
            ACTION_TYPEHASH=action_typehash,  # from Protocol Constants table in docs.derive_perpetual.xyz
        )
        try:
            action.sign(self.session_key_wallet.key)
        except Exception as e:
            raise Exception(f"Error signing action: {e}")

        return action.to_json()

    def header_for_authentication(self) -> Dict[str, str]:
        timestamp = str(self.utc_now_ms())
        signature = to_0x_hex(self._w3.eth.account.sign_message(
            encode_defunct(text=timestamp), private_key=self._api_secret
        ).signature)
        payload = {}

        payload["accept"] = 'application/json'
        payload["X-LyraWallet"] = self._api_key
        payload["X-LyraTimestamp"] = timestamp
        payload["X-LyraSignature"] = signature
        return payload

    @staticmethod
    def utc_now_ms() -> int:
        return int(datetime.now(timezone.utc).timestamp() * 1000)



from decimal import Decimal
from typing import Dict, Union

from pydantic import ConfigDict, Field, field_validator

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.connector.exchange.injective_v2.injective_v2_utils import (
    ACCOUNT_MODES,
    FEE_CALCULATOR_MODES,
    NETWORK_MODES,
    InjectiveMainnetNetworkMode,
    InjectiveMessageBasedTransactionFeeCalculatorMode,
    InjectiveReadOnlyAccountMode,
)
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = False
EXAMPLE_PAIR = "INJ-USDT"

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0"),
    taker_percent_fee_decimal=Decimal("0"),
)


class InjectiveConfigMap(BaseConnectorConfigMap):
    # Setting a default dummy configuration to allow the bot to create a dummy instance to fetch all trading pairs
    connector: str = "injective_v2_perpetual"
    receive_connector_configuration: bool = Field(default=True)
    network: Union[tuple(NETWORK_MODES.values())] = Field(
        default=InjectiveMainnetNetworkMode(),
        json_schema_extra={
            "prompt": lambda cm: f"Select the network ({'/'.join(list(NETWORK_MODES.keys()))})",
            "prompt_on_new": True,
        }
    )
    account_type: Union[tuple(ACCOUNT_MODES.values())] = Field(
        default=InjectiveReadOnlyAccountMode(),
        json_schema_extra={
            "prompt": lambda cm: f"Select the type of account ({'/'.join(list(ACCOUNT_MODES.keys()))})",
            "prompt_on_new": True
        },
    )
    fee_calculator: Union[tuple(FEE_CALCULATOR_MODES.values())] = Field(
        default=InjectiveMessageBasedTransactionFeeCalculatorMode(),
        discriminator="name",
        json_schema_extra={
            "prompt": lambda cm: f"Select the fee calculator ({'/'.join(list(FEE_CALCULATOR_MODES.keys()))})",
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="injective_v2_perpetual")

    @field_validator("network", mode="before")
    @classmethod
    def validate_network(cls, v: Union[(str, Dict) + tuple(NETWORK_MODES.values())]):
        if isinstance(v, tuple(NETWORK_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in NETWORK_MODES:
            raise ValueError(
                f"Invalid network, please choose a value from {list(NETWORK_MODES.keys())}."
            )
        else:
            sub_model = NETWORK_MODES[v].model_construct()
        return sub_model

    @field_validator("account_type", mode="before")
    @classmethod
    def validate_account_type(cls, v: Union[(str, Dict) + tuple(ACCOUNT_MODES.values())]):
        if isinstance(v, tuple(ACCOUNT_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in ACCOUNT_MODES:
            raise ValueError(
                f"Invalid account type, please choose a value from {list(ACCOUNT_MODES.keys())}."
            )
        else:
            sub_model = ACCOUNT_MODES[v].model_construct()
        return sub_model

    @field_validator("fee_calculator", mode="before")
    @classmethod
    def validate_fee_calculator(cls, v: Union[(str, Dict) + tuple(FEE_CALCULATOR_MODES.values())]):
        if isinstance(v, tuple(FEE_CALCULATOR_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in FEE_CALCULATOR_MODES:
            raise ValueError(
                f"Invalid fee calculator, please choose a value from {list(FEE_CALCULATOR_MODES.keys())}."
            )
        else:
            sub_model = FEE_CALCULATOR_MODES[v].model_construct()
        return sub_model

    def create_data_source(self):
        return self.account_type.create_data_source(
            network=self.network.network(),
            rate_limits=self.network.rate_limits(),
            fee_calculator_mode=self.fee_calculator,
        )


KEYS = InjectiveConfigMap.model_construct()



#!/usr/bin/env python
import os
from typing import TYPE_CHECKING, Any, List, Optional

from hummingbot.client.command.gateway_api_manager import begin_placeholder_mode
from hummingbot.core.gateway.gateway_http_client import GatewayStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.gateway_config_utils import build_config_dict_display

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


def ensure_gateway_online(func):
    def wrapper(self, *args, **kwargs):
        if self.trading_core.gateway_monitor.gateway_status is GatewayStatus.OFFLINE:
            self.logger().error("Gateway is offline")
            return
        return func(self, *args, **kwargs)
    return wrapper


class GatewayConfigCommand:
    """Commands for managing gateway configuration."""

    @ensure_gateway_online
    def gateway_config(self, namespace: str = None, action: str = None, args: List[str] = None):
        """
        Gateway configuration management.
        Usage:
            gateway config [namespace]                       - Show configuration for namespace
            gateway config <namespace> update                 - Update configuration (interactive)
            gateway config <namespace> update <path> <value>  - Update configuration (direct)
        """
        if args is None:
            args = []

        if namespace is None:
            # Show help when no namespace is provided
            self.notify("\nUsage:")
            self.notify("  gateway config [namespace]                       - Show configuration")
            self.notify("  gateway config <namespace> update                 - Update configuration (interactive)")
            self.notify("  gateway config <namespace> update <path> <value>  - Update configuration (direct)")
            self.notify("\nExamples:")
            self.notify("  gateway config ethereum-mainnet")
            self.notify("  gateway config uniswap")
            self.notify("  gateway config ethereum-mainnet update")
            self.notify("  gateway config solana-mainnet update nodeURL https://api.mainnet-beta.solana.com")
        elif action is None:
            # Format: gateway config <namespace>
            # Show configuration for the specified namespace
            safe_ensure_future(
                GatewayConfigCommand._show_gateway_configuration(self, namespace=namespace),
                loop=self.ev_loop
            )
        elif action == "update":
            if len(args) >= 2:
                # Non-interactive mode: gateway config <namespace> update <path> <value>
                path = args[0]
                # Join remaining args as the value (in case value contains spaces)
                value = " ".join(args[1:])
                safe_ensure_future(
                    GatewayConfigCommand._update_gateway_configuration_direct(self, namespace, path, value),
                    loop=self.ev_loop
                )
            else:
                # Interactive mode: gateway config <namespace> update
                safe_ensure_future(
                    GatewayConfigCommand._update_gateway_configuration_interactive(self, namespace),
                    loop=self.ev_loop
                )
        else:
            # If action is not "update", it might be a namespace typo
            self.notify(f"\nError: Invalid action '{action}'. Use 'update' to modify configuration.")
            self.notify("\nUsage:")
            self.notify("  gateway config <namespace>                       - Show configuration")
            self.notify("  gateway config <namespace> update                 - Update configuration (interactive)")
            self.notify("  gateway config <namespace> update <path> <value>  - Update configuration (direct)")

    async def _show_gateway_configuration(
        self,  # type: HummingbotApplication
        namespace: Optional[str] = None,
    ):
        """Show gateway configuration for a namespace."""
        host = self.client_config_map.gateway.gateway_api_host
        port = self.client_config_map.gateway.gateway_api_port
        try:
            config_dict = await self._get_gateway_instance().get_configuration(namespace=namespace)
            # Format the title
            title_parts = ["Gateway Configuration"]
            if namespace:
                title_parts.append(f"namespace: {namespace}")
            title = f"\n{' - '.join(title_parts)}:"

            self.notify(title)
            lines = []
            build_config_dict_display(lines, config_dict)
            self.notify("\n".join(lines))

        except Exception:
            remote_host = ':'.join([host, port])
            self.notify(f"\nError: Connection to Gateway {remote_host} failed")

    async def _update_gateway_configuration(
        self,  # type: HummingbotApplication
        namespace: str,
        key: str,
        value: Any
    ):
        """Update a single gateway configuration value."""
        try:
            response = await self._get_gateway_instance().update_config(
                namespace=namespace,
                path=key,
                value=value
            )
            self.notify(response["message"])
        except Exception:
            self.notify(
                "\nError: Gateway configuration update failed. See log file for more details."
            )

    async def _update_gateway_configuration_direct(
        self,  # type: HummingbotApplication
        namespace: str,
        path: str,
        value: str
    ):
        """Direct mode for gateway config update with validation."""
        try:
            # Get the current configuration to validate the path
            config_dict = await self._get_gateway_instance().get_configuration(namespace=namespace)

            if not config_dict:
                self.notify(f"No configuration found for namespace: {namespace}")
                return

            # Get available config keys
            config_keys = list(config_dict.keys())

            # Validate the path
            if path not in config_keys:
                self.notify(f"\nError: Invalid configuration path '{path}'")
                self.notify(f"Valid paths are: {', '.join(config_keys)}")
                return

            # Get current value for type validation
            current_value = config_dict.get(path)
            self.notify(f"\nUpdating {namespace}.{path}")
            self.notify(f"Current value: {current_value}")
            self.notify(f"New value: {value}")

            # Validate the value based on the current value type
            validated_value = await GatewayConfigCommand._validate_config_value(
                self,
                path,
                value,
                current_value,
                namespace
            )

            if validated_value is None:
                return

            # Update the configuration
            await GatewayConfigCommand._update_gateway_configuration(
                self,
                namespace,
                path,
                validated_value
            )

        except Exception as e:
            self.notify(f"Error updating configuration: {str(e)}")

    async def _update_gateway_configuration_interactive(
        self,  # type: HummingbotApplication
        namespace: str
    ):
        """Interactive mode for gateway config update with path validation."""
        try:
            # First get the current configuration to show available paths
            config_dict = await self._get_gateway_instance().get_configuration(namespace=namespace)

            if not config_dict:
                self.notify(f"No configuration found for namespace: {namespace}")
                return

            # Display current configuration
            self.notify(f"\nCurrent configuration for {namespace}:")
            lines = []
            build_config_dict_display(lines, config_dict)
            self.notify("\n".join(lines))

            # Get available config keys
            config_keys = list(config_dict.keys())

            # Enter interactive mode
            with begin_placeholder_mode(self):
                try:
                    # Update completer's config path options
                    if hasattr(self.app.input_field.completer, '_gateway_config_path_options'):
                        self.app.input_field.completer._gateway_config_path_options = config_keys

                    # Loop to allow retry on invalid path
                    while True:
                        # Prompt for path
                        self.notify(f"\nAvailable configuration paths: {', '.join(config_keys)}")
                        path = await self.app.prompt(prompt="Enter configuration path (or 'exit' to cancel): ")

                        if self.app.to_stop_config or not path or path.lower() == 'exit':
                            self.notify("Configuration update cancelled")
                            return

                        # Validate the path
                        if path not in config_keys:
                            self.notify(f"\nError: Invalid configuration path '{path}'")
                            self.notify(f"Valid paths are: {', '.join(config_keys)}")
                            self.notify("Please try again.")
                            continue  # Allow retry

                        # Valid path, break the loop
                        break

                    # Show current value
                    current_value = config_dict.get(path, "Not found")
                    self.notify(f"\nCurrent value for '{path}': {current_value}")

                    # Loop to allow retry on invalid value
                    while True:
                        # Prompt for new value
                        value = await self.app.prompt(prompt="Enter new value (or 'exit' to cancel): ")

                        if self.app.to_stop_config or not value or value.lower() == 'exit':
                            self.notify("Configuration update cancelled")
                            return

                        # Validate the value based on the current value type
                        validated_value = await GatewayConfigCommand._validate_config_value(
                            self,
                            path,
                            value,
                            current_value,
                            namespace
                        )

                        if validated_value is None:
                            self.notify("Please try again.")
                            continue  # Allow retry

                        # Valid value, break the loop
                        break

                    # Update the configuration
                    await GatewayConfigCommand._update_gateway_configuration(
                        self,
                        namespace,
                        path,
                        validated_value
                    )

                finally:
                    self.placeholder_mode = False
                    self.app.hide_input = False
                    self.app.change_prompt(prompt=">>> ")

        except Exception as e:
            self.notify(f"Error in interactive config update: {str(e)}")

    async def _validate_config_value(
        self,  # type: HummingbotApplication
        path: str,
        value: str,
        current_value: Any,
        namespace: str = None
    ) -> Optional[Any]:
        """
        Validate and convert the config value based on the current value type.
        Also performs special validation for path values and network values.
        """
        try:
            # Special validation for path-like configuration keys
            path_keywords = ['path', 'dir', 'directory', 'folder', 'location']
            is_path_config = any(keyword in path.lower() for keyword in path_keywords)

            # Type conversion based on current value
            if isinstance(current_value, bool):
                # Boolean conversion
                if value.lower() in ['true', 'yes', '1']:
                    return True
                elif value.lower() in ['false', 'no', '0']:
                    return False
                else:
                    self.notify(f"Error: Expected boolean value (true/false), got '{value}'")
                    return None

            elif isinstance(current_value, (int, float)):
                # Numeric conversion - accept both int and float values
                # This allows reverting from integer to decimal values
                try:
                    parsed = float(value)
                    # Return int if the value is a whole number and current is int
                    if isinstance(current_value, int) and parsed == int(parsed):
                        return int(parsed)
                    return parsed
                except ValueError:
                    self.notify(f"Error: Expected numeric value, got '{value}'")
                    return None

            elif isinstance(current_value, str):
                # String value - check if it's a path
                if is_path_config:
                    # Validate path
                    expanded_path = os.path.expanduser(value)
                    if not os.path.exists(expanded_path):
                        self.notify(f"\nError: Path does not exist: {expanded_path}")
                        self.notify("Please enter a valid path.")
                        return None
                    # Return the original value (not expanded) as Gateway handles expansion
                    return value
                elif path.lower() == "defaultnetwork" and namespace:
                    # Special validation for defaultNetwork - must be a valid network for the chain
                    # Await the async validation
                    available_networks = await self._get_gateway_instance().get_available_networks_for_chain(

                        namespace  # namespace is the chain name
                    )

                    if available_networks and value not in available_networks:
                        self.notify(f"\nError: Invalid network '{value}' for {namespace}")
                        self.notify("Valid networks are: " + ", ".join(available_networks))
                        return None

                    return value
                else:
                    # Regular string
                    return value

            elif isinstance(current_value, list):
                # List conversion - try to parse as comma-separated values
                if value.startswith('[') and value.endswith(']'):
                    # JSON-style list
                    import json
                    try:
                        return json.loads(value)
                    except json.JSONDecodeError:
                        self.notify("Error: Invalid list format. Use JSON format like: ['item1', 'item2']")
                        return None
                else:
                    # Comma-separated values
                    return [item.strip() for item in value.split(',')]

            else:
                # Unknown type - return as string
                return value

        except Exception as e:
            self.notify(f"Error validating value: {str(e)}")
            return None



from hummingbot.core.data_type.common import PriceType
from hummingbot.connector.exchange_base import ExchangeBase
from decimal import Decimal
from .asset_price_delegate cimport AssetPriceDelegate

cdef class OrderBookAssetPriceDelegate(AssetPriceDelegate):
    def __init__(self, market: ExchangeBase, trading_pair: str):
        super().__init__()
        self._market = market
        self._trading_pair = trading_pair

    cdef object c_get_mid_price(self):
        return (self._market.c_get_price(self._trading_pair, True) +
                self._market.c_get_price(self._trading_pair, False))/Decimal('2')

    @property
    def ready(self) -> bool:
        return self._market.ready

    def get_price_by_type(self, price_type: PriceType) -> Decimal:
        return self._market.get_price_by_type(self._trading_pair, price_type)

    @property
    def market(self) -> ExchangeBase:
        return self._market

    @property
    def trading_pair(self) -> str:
        return self._trading_pair



import asyncio
import platform
import threading
from typing import TYPE_CHECKING, Callable, List, Optional, Set

import hummingbot.client.settings as settings
from hummingbot import init_logging
from hummingbot.client.command.gateway_api_manager import GatewayChainApiManager
from hummingbot.client.config.config_validators import validate_bool
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.exceptions import OracleRateUnavailable

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401

GATEWAY_READY_TIMEOUT = 300  # seconds


class StartCommand(GatewayChainApiManager):
    _in_start_check: bool = False

    async def _run_clock(self):
        with self.trading_core.clock as clock:
            await clock.run()

    async def wait_till_ready(self,  # type: HummingbotApplication
                              func: Callable, *args, **kwargs):
        while True:
            all_ready = all([market.ready for market in self.trading_core.markets.values()])
            if not all_ready:
                await asyncio.sleep(0.5)
            else:
                return func(*args, **kwargs)

    def _strategy_uses_gateway_connector(self, required_exchanges: Set[str]) -> bool:
        exchange_settings: List[settings.ConnectorSetting] = [
            settings.AllConnectorSettings.get_connector_settings().get(e, None)
            for e in required_exchanges
        ]
        return any([s.uses_gateway_generic_connector()
                    for s in exchange_settings])

    def start(self,  # type: HummingbotApplication
              log_level: Optional[str] = None,
              script: Optional[str] = None,
              conf: Optional[str] = None,
              is_quickstart: Optional[bool] = False):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.start, log_level, script)
            return
        safe_ensure_future(self.start_check(log_level, script, conf, is_quickstart), loop=self.ev_loop)

    async def start_check(self,  # type: HummingbotApplication
                          log_level: Optional[str] = None,
                          script: Optional[str] = None,
                          conf: Optional[str] = None,
                          is_quickstart: Optional[bool] = False):

        if self._in_start_check or (
                self.trading_core.strategy_task is not None and not self.trading_core.strategy_task.done()):
            self.notify('The bot is already running - please run "stop" first')
            return

        self._in_start_check = True

        if settings.required_rate_oracle:
            # If the strategy to run requires using the rate oracle to find FX rates, validate there is a rate for
            # each configured token pair
            if not (await self.confirm_oracle_conversion_rate()):
                self.notify("The strategy failed to start.")
                self._in_start_check = False
                return

        if self.strategy_file_name and self.trading_core.strategy_name and is_quickstart:
            if self._strategy_uses_gateway_connector(settings.required_exchanges):
                try:
                    await asyncio.wait_for(self.trading_core.gateway_monitor.ready_event.wait(), timeout=GATEWAY_READY_TIMEOUT)
                except asyncio.TimeoutError:
                    self.notify(
                        f"TimeoutError waiting for gateway service to go online... Please ensure Gateway is configured correctly."
                        f"Unable to start strategy {self.trading_core.strategy_name}. ")
                    self._in_start_check = False
                    self.trading_core.strategy_name = None
                    self.strategy_file_name = None
                    raise

        if script:
            file_name = script.split(".")[0]
            self.trading_core.strategy_name = file_name
            self.strategy_file_name = conf if conf else file_name
        elif not await self.status_check_all(notify_success=False):
            self.notify("Status checks failed. Start aborted.")
            self._in_start_check = False
            return
        init_logging("hummingbot_logs.yml",
                     self.client_config_map,
                     override_log_level=log_level.upper() if log_level else None,
                     strategy_file_path=self.strategy_file_name)

        # If macOS, disable App Nap.
        if platform.system() == "Darwin":
            import appnope
            appnope.nope()

        self._initialize_notifiers()

        # Delegate strategy initialization to trading_core
        try:
            strategy_config = None
            if self.trading_core.is_script_strategy(self.trading_core.strategy_name):
                if self.strategy_file_name and self.strategy_file_name != self.trading_core.strategy_name:
                    strategy_config = self.strategy_file_name

            success = await self.trading_core.start_strategy(
                self.trading_core.strategy_name,
                strategy_config,
                self.strategy_file_name
            )
            if not success:
                self._in_start_check = False
                self.trading_core.strategy_name = None
                self.strategy_file_name = None
                self.notify("Invalid strategy. Start aborted.")
                return
        except Exception as e:
            self._in_start_check = False
            self.trading_core.strategy_name = None
            self.strategy_file_name = None
            self.notify(f"Invalid strategy. Start aborted {e}.")
            raise

        if any([str(exchange).endswith("paper_trade") for exchange in settings.required_exchanges]):
            self.notify("\nPaper Trading Active: All orders are simulated and no real orders are placed.")

        self.notify(f"\nStatus check complete. Strategy '{self.trading_core.strategy_name}' started successfully.")
        self._in_start_check = False

        # Patch MQTT loggers if MQTT is available
        if self._mqtt:
            self._mqtt.patch_loggers()
            self._mqtt.start_market_events_fw()

    async def confirm_oracle_conversion_rate(self,  # type: HummingbotApplication
                                             ) -> bool:
        try:
            result = False
            self.app.clear_input()
            self.placeholder_mode = True
            self.app.hide_input = True
            for pair in settings.rate_oracle_pairs:
                msg = await self.oracle_rate_msg(pair)
                self.notify("\nRate Oracle:\n" + msg)
            config = ConfigVar(key="confirm_oracle_use",
                               type_str="bool",
                               prompt="Please confirm to proceed if the above oracle source and rates are correct for "
                                      "this strategy (Yes/No)  >>> ",
                               required_if=lambda: True,
                               validator=lambda v: validate_bool(v))
            await self.prompt_a_config_legacy(config)
            if config.value:
                result = True
        except OracleRateUnavailable:
            self.notify("Oracle rate is not available.")
        finally:
            self.placeholder_mode = False
            self.app.hide_input = False
            self.app.change_prompt(prompt=">>> ")
        return result
